from builtins import zip

__author__ = "giacomov"

import collections
import os
import warnings
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import scipy.integrate

from astromodels.core.memoization import use_astromodels_memoization
from astromodels.core.my_yaml import my_yaml
from astromodels.core.parameter import IndependentVariable, Parameter
from astromodels.core.tree import DuplicatedNode, Node
from astromodels.functions.function import get_function
from astromodels.sources import (ExtendedSource, ParticleSource, PointSource,
                                 Source, SourceType)
from astromodels.utils.disk_usage import disk_usage
from astromodels.utils.logging import setup_logger
from astromodels.utils.long_path_formatter import long_path_formatter

log = setup_logger(__name__)


class ModelFileExists(IOError):
    pass


class InvalidInput(ValueError):
    pass


class CannotWriteModel(IOError):
    def __init__(self, directory, message):
        # Add a report on disk usage to the message

        free_space = disk_usage(directory).free

        message += "\nFree space on the file system hosting %s was %.2f Mbytes" % (
            directory,
            free_space / 1024.0 / 1024.0,
        )

        super(CannotWriteModel, self).__init__(message)


class ModelInternalError(ValueError):

    pass


class Model(Node):
    def __init__(self, *sources):

        # Setup the node, using the special name '__root__' to indicate that this is the root of the tree

        super(Model, self).__init__("__root__")

        # Dictionary to keep point sources

        self._point_sources: Dict[str, PointSource] = collections.OrderedDict()

        # Dictionary to keep extended sources

        self._extended_sources: Dict[
            str, ExtendedSource] = collections.OrderedDict()

        # Dictionary to keep particle sources

        self._particle_sources: Dict[
            str, ParticleSource] = collections.OrderedDict()

        # Loop over the provided sources and process them

        for source in sources:

            self._add_source(source)

        # Now make the list of all the existing parameters

        self._update_parameters()

        # This controls the verbosity of the display
        self._complete_display = False

        # This will keep track of independent variables (if any)
        self._independent_variables = {}

    def _add_source(
            self, source: Union[PointSource, ExtendedSource,
                                ParticleSource]) -> None:
        """
        Remember to call _update_parameters after this!
        :param source:
        :return:
        """

        try:

            self._add_child(source)

        except AttributeError:

            if isinstance(source, Source):

                log.error(
                    "More than one source with the name '%s'. You cannot use the same name for multiple "
                    "sources" % source.name)

                raise DuplicatedNode()

            else:  # pragma: no cover

                raise

        # Now see if this is a point or extended source, and add them to the
        # appropriate dictionary

        if source.source_type == SourceType.POINT_SOURCE:

            self._point_sources[source.name] = source

        elif source.source_type == SourceType.EXTENDED_SOURCE:

            self._extended_sources[source.name] = source

        elif source.source_type == SourceType.PARTICLE_SOURCE:

            self._particle_sources[source.name] = source

        else:  # pragma: no cover

            raise InvalidInput(
                "Input sources must be either a point source or an extended source"
            )

    def _remove_source(self, source_name: str):
        """
        Remember to call _update_parameters after this
        :param source_name:
        :return:
        """

        if not source_name in self.sources:

            log.error(f"Source {source_name} is not part of the current model")

            raise AssertionError()

        source = self.sources.pop(source_name)

        if source.source_type == SourceType.POINT_SOURCE:

            self._point_sources.pop(source.name)

        elif source.source_type == SourceType.EXTENDED_SOURCE:

            self._extended_sources.pop(source.name)

        elif source.source_type == SourceType.PARTICLE_SOURCE:

            self._particle_sources.pop(source.name)

        self._remove_child(source_name)

    def _find_parameters(self, node) -> Dict[str, Parameter]:

        instances = collections.OrderedDict()

        for child in node._get_children():

            if isinstance(child, Parameter):

                path = child._get_path()

                instances[path] = child

                for sub_child in child._get_children():

                    instances.update(self._find_parameters(sub_child))

            else:

                instances.update(self._find_parameters(child))

        return instances

    def _update_parameters(self) -> None:

        self._parameters: Dict[str, Parameter] = self._find_parameters(self)

    @property
    def parameters(self) -> Dict[str, Parameter]:
        """
        Return a dictionary with all parameters

        :return: dictionary of parameters
        """
        self._update_parameters()

        return self._parameters

    @property
    def free_parameters(self) -> Dict[str, Parameter]:
        """
        Get a dictionary with all the free parameters in this model

        :return: dictionary of free parameters
        """

        # Refresh the list

        self._update_parameters()

        # Filter selecting only free parameters

        free_parameters_dictionary: Dict[
            str, Parameter] = collections.OrderedDict()

        for parameter_name, parameter in list(self._parameters.items()):

            if parameter.free:

                free_parameters_dictionary[parameter_name] = parameter

        return free_parameters_dictionary

    @property
    def linked_parameters(self) -> Dict[str, Parameter]:
        """
        Get a dictionary with all parameters in this model in a linked status. A parameter is in a linked status
        if it is linked to another parameter (i.e. it is forced to have the same value of the other parameter), or
        if it is linked with another parameter or an independent variable through a law.

        :return: dictionary of linked parameters
        """

        # Refresh the list

        self._update_parameters()

        # Filter selecting only free parameters

        linked_parameter_dictionary: Dict[
            str, Parameter] = collections.OrderedDict()

        for parameter_name, parameter in list(self._parameters.items()):

            if parameter.has_auxiliary_variable():

                linked_parameter_dictionary[parameter_name] = parameter

        return linked_parameter_dictionary

    def set_free_parameters(self, values: Iterable[float]) -> None:
        """
        Set the free parameters in the model to the provided values.

        NOTE: of course, order matters

        :param values: a list of new values
        :return: None
        """

        if not len(values) == len(self.free_parameters):

            log.error(
                f"tried to pass {len(values)} parameters but need {len(self.free_parameters)}"
            )

            raise AssertionError()

        for parameter, this_value in zip(list(self.free_parameters.values()),
                                         values):

            parameter.value = this_value

    def __getitem__(self, path: str):
        """
        Get a parameter from a path like "source_1.component.powerlaw.logK". This might be useful in certain
        context, although in an interactive analysis there is no reason to use this.

        :param path: the address of the parameter
        :return: the parameter
        """

        return self._get_child_from_path(path)

    def __contains__(self, path: str) -> bool:
        """
        This allows the model to be used with the "in" operator, like;

        > if 'myparameter' in model:
        >    print("Myparameter is contained in the model")

        :param path: the parameter to look for
        :return:
        """

        try:

            _ = self._get_child_from_path(path)

        except (AttributeError, KeyError, TypeError):

            return False

        else:

            return True

    def __iter__(self):
        """
        This allows the model to be iterated on, like in:

        for parameter in model:
            ...

        NOTE: this will iterate over *all* parameters in the model, also those that are not free (and thus are not
        normally displayed). If you need to operate only on free parameters, just check if they are free within
        the loop or use the .free_parameters dictionary directly

        :return: iterator
        """

        for parameter in self.parameters:

            yield self.parameters[parameter]

    @property
    def point_sources(self) -> Dict[str, PointSource]:
        """
        Returns the dictionary of all defined point sources

        :return: collections.OrderedDict()
        """
        return self._point_sources

    @property
    def extended_sources(self) -> Dict[str, ExtendedSource]:
        """
        Returns the dictionary of all defined extended sources

        :return: collections.OrderedDict()

        """
        return self._extended_sources

    @property
    def particle_sources(self) -> Dict[str, ParticleSource]:
        """
        Returns the dictionary of all defined particle sources

        :return: collections.OrderedDict()

        """
        return self._particle_sources

    @property
    def sources(
            self
    ) -> Dict[str, Union[PointSource, ExtendedSource, ParticleSource]]:
        """
        Returns a dictionary containing all defined sources (of any kind)

        :return: collections.OrderedDict()

        """

        sources: Dict[str, Union[PointSource, ExtendedSource,
                                 ParticleSource]] = collections.OrderedDict()

        for d in (self.point_sources, self.extended_sources,
                  self.particle_sources):

            sources.update(d)

        return sources

    def add_source(
        self, new_source: Union[PointSource, ExtendedSource,
                                ParticleSource]) -> None:
        """
        Add the provided source to the model

        :param new_source: the new source to be added (an instance of PointSource, ExtendedSource or ParticleSource)
        :return: (none)
        """

        self._add_source(new_source)

        self._update_parameters()

    def remove_source(self, source_name: str) -> None:
        """
        Returns a new model with the provided source removed from the current model. Any parameters linked to the source to be removed are automatically unlinked.

        :param source_name: the name of the source to be removed
        :return: a new Model instance without the source
        """

        self.unlink_all_from_source(source_name, warn=True)

        self._remove_source(source_name)

        self._update_parameters()

    def unlink_all_from_source(self,
                               source_name: str,
                               warn: bool = False) -> None:
        """
        Unlink all parameters of the current model that are linked to a parameter of a given source.
        To be called before removing a source from the model.

        :param source_name: the name of the source to which to remove all links
        :param warn: If True, prints a warning if any parameters were unlinked.
        """

        tempmodel = Model(self[source_name])
        unlinked_parameters = collections.OrderedDict()

        for par in self.linked_parameters.values():

            target = par._aux_variable['variable']

            if target.path in tempmodel:

                unlinked_parameters[par.name] = par
                self.unlink(par)

        if warn and unlinked_parameters:

            log.warning(
                "The following %d parameters that were linked to source %s have been automatically un-linked: %s"
                % (len(unlinked_parameters), source_name,
                   [p.path for p in unlinked_parameters.values()]),
                RuntimeWarning)

    def add_independent_variable(self, variable: IndependentVariable) -> None:
        """
        Add a global independent variable to this model, such as time.

        :param variable: an IndependentVariable instance
        :return: none
        """

        if not isinstance(variable, IndependentVariable):
            log.error("Variable must be an instance of IndependentVariable")

            raise AssertionError()

        if self._has_child(variable.name):

            self._remove_child(variable.name)

        self._add_child(variable)

        # Add also to the list of independent variables
        self._independent_variables[variable.name] = variable

    def remove_independent_variable(self, variable_name: str) -> None:
        """
        Remove an independent variable which was added with add_independent_variable

        :param variable_name: name of variable to remove
        :return:
        """

        self._remove_child(variable_name)

        # Remove also from the list of independent variables
        self._independent_variables.pop(variable_name)

    def add_external_parameter(self, parameter: Parameter) -> None:
        """
        Add a parameter that comes from something other than a function, to the model.

        :param parameter: a Parameter instance
        :return: none
        """

        if not isinstance(parameter, Parameter):

            log.error("Variable must be an instance of IndependentVariable")

            raise AssertionError()

        if self._has_child(parameter.name):

            # Remove it from the children only if it is a Parameter instance, otherwise don't, which will
            # make the _add_child call fail (which is the expected behaviour! You shouldn't call two children
            # with the same name)

            if isinstance(self._get_child(parameter.name), Parameter):

                log.warning(
                    "External parameter %s already exist in the model. Overwriting it..."
                    % parameter.name)

                self._remove_child(parameter.name)

        # This will fail if another node with the same name is already in the model

        self._add_child(parameter)

    def remove_external_parameter(self, parameter_name: str) -> None:
        """
        Remove an external parameter which was added with add_external_parameter

        :param variable_name: name of parameter to remove
        :return:
        """

        self._remove_child(parameter_name)

    def link(self, parameter_1, parameter_2, link_function=None) -> None:
        """
        Link the value of the provided parameters through the provided function (identity is the default, i.e.,
        parameter_1 = parameter_2).

        :param parameter_1: the first parameter;can be either a single parameter or a list of prarameters
        :param parameter_2: the second parameter
        :param link_function: a function instance. If not provided, the identity function will be used by default.
        Otherwise, this link will be set: parameter_1 = link_function(parameter_2)
        :return: (none)
        """
        if not isinstance(parameter_1, list):
            # Make a list of one element
            parameter_1_list = [parameter_1]
        else:
            # Make a copy to avoid tampering with the input
            parameter_1_list = list(parameter_1)

        for param_1 in parameter_1_list:
            if not param_1.path in self:

                log.error( "Parameter %s is not contained in this model" % param_1.path )

                raise AssertionError()
                
        if not parameter_2.path in self:

            log.error( "Parameter %s is not contained in this model" % parameter_2.path )

            raise AssertionError()
            
        if link_function is None:
            # Use the Line function by default, with both parameters fixed so that the two
            # parameters to be linked will vary together
            link_function = get_function("Line")

            link_function.a.value = 0
            link_function.a.fix = True

            link_function.b.value = 1
            link_function.b.fix = True

        for param_1 in parameter_1_list:
            param_1.add_auxiliary_variable(parameter_2, link_function)
            # Now set the units of the link function
            link_function.set_units(parameter_2.unit, param_1.unit)

    def unlink(self, parameter: Parameter) -> None:
        """
        Sets free one or more parameters which have been linked previously

        :param parameter: the parameter to be set free, can also be a list of parameters
        :return: (none)
        """

        if not isinstance(parameter, list):
            # Make a list of one element
            parameter_list = [parameter]
        else:
            # Make a copy to avoid tampering with the input
            parameter_list = list(parameter)

        for param in parameter_list:
            if param.has_auxiliary_variable():
                param.remove_auxiliary_variable()

            else:

                with warnings.catch_warnings():

                    warnings.simplefilter("always", RuntimeWarning)

                    log.warning(
                        "Parameter %s has no link to be removed." % param.path
                       
                    )

    def display(self, complete: bool=False) -> None:
        """
        Display information about the point source.

        :param complete : if True, displays also information on fixed parameters
        :return: (none)
        """

        # Switch on the complete display flag
        self._complete_display = bool(complete)

        # This will automatically choose the best representation among repr and repr_html

        super(Model, self).display()

        # Go back to default

        self._complete_display = False

    def _repr__base(self, rich_output=False):

        if rich_output:

            new_line = "<br>"

        else:

            new_line = "\n"

        # Table with the summary of the various kind of sources
        sources_summary = pd.DataFrame.from_dict(
            collections.OrderedDict(
                [
                    ("Point sources", [self.get_number_of_point_sources()]),
                    ("Extended sources", [self.get_number_of_extended_sources()]),
                    ("Particle sources", [self.get_number_of_particle_sources()]),
                ]
            ),
            columns=["N"],
            orient="index",
        )

        # These properties traverse the whole tree everytime, so let's cache their results here
        parameters = self.parameters
        free_parameters = self.free_parameters
        linked_parameters = self.linked_parameters

        # Summary of free parameters
        if len(free_parameters) > 0:

            parameter_dict = collections.OrderedDict()

            for parameter_name, parameter in list(free_parameters.items()):
                # Generate table with only a minimal set of info

                # Generate table with only a minimal set of info
                if rich_output:

                    this_name = long_path_formatter(parameter_name, 70)

                else:

                    # In a terminal we need to use less characters

                    this_name = long_path_formatter(parameter_name, 40)

                d = parameter.to_dict()
                parameter_dict[this_name] = collections.OrderedDict()

                for key in ["value", "unit", "min_value", "max_value"]:

                    parameter_dict[this_name][key] = d[key]

            free_parameters_summary = pd.DataFrame.from_dict(parameter_dict).T

            # Re-order it
            free_parameters_summary = free_parameters_summary[
                ["value", "min_value", "max_value", "unit"]
            ]

        else:

            free_parameters_summary = pd.DataFrame()

        if len(parameters) - len(free_parameters) - len(linked_parameters) > 0:

            fixed_parameter_dict = collections.OrderedDict()

            for parameter_name, parameter in list(parameters.items()):

                if parameter.free or parameter_name in linked_parameters:

                    continue

                # Generate table with only a minimal set of info
                if rich_output:

                    this_name = long_path_formatter(parameter_name, 70)

                else:

                    # In a terminal we need to use less characters

                    this_name = long_path_formatter(parameter_name, 40)

                d = parameter.to_dict()
                fixed_parameter_dict[this_name] = collections.OrderedDict()

                for key in ["value", "unit", "min_value", "max_value"]:

                    fixed_parameter_dict[this_name][key] = d[key]

            fixed_parameters_summary = pd.DataFrame.from_dict(fixed_parameter_dict).T

            # Re-order it
            fixed_parameters_summary = fixed_parameters_summary[
                ["value", "min_value", "max_value", "unit"]
            ]

        else:

            fixed_parameters_summary = pd.DataFrame()

        # Summary of linked parameters

        linked_frames = []

        if linked_parameters:

            for parameter_name, parameter in list(linked_parameters.items()):

                parameter_dict = collections.OrderedDict()

                # Generate table with only a minimal set of info

                variable, law = parameter.auxiliary_variable

                this_dict = collections.OrderedDict()

                this_dict["linked to"] = variable.path
                this_dict["function"] = law.name
                this_dict["current value"] = parameter.value
                this_dict["unit"] = parameter.unit

                parameter_dict[parameter_name] = this_dict

                this_parameter_frame = pd.DataFrame.from_dict(parameter_dict)

                linked_frames.append(this_parameter_frame)

        else:

            # No linked parameters

            pass

        empty_frame = "(none)%s" % new_line

        # Independent variables

        independent_v_frames = []

        if self._independent_variables:

            for variable_name, variable_instance in list(
                self._independent_variables.items()
            ):

                v_dict = collections.OrderedDict()

                # Generate table with only a minimal set of info

                this_dict = collections.OrderedDict()

                this_dict["current value"] = variable_instance.value
                this_dict["unit"] = variable_instance.unit

                v_dict[variable_name] = this_dict

                this_v_frame = pd.DataFrame.from_dict(v_dict)

                independent_v_frames.append(this_v_frame)

        else:

            # No independent variables

            pass

        if rich_output:

            source_summary_representation = sources_summary._repr_html_()

            if free_parameters_summary.empty:

                free_parameters_representation = empty_frame

            else:

                free_parameters_representation = free_parameters_summary._repr_html_()

            if len(linked_frames) == 0:

                linked_summary_representation = empty_frame

            else:

                linked_summary_representation = ""

                for linked_frame in linked_frames:

                    linked_summary_representation += linked_frame._repr_html_()
                    linked_summary_representation += new_line

            if len(independent_v_frames) == 0:

                independent_v_representation = empty_frame

            else:

                independent_v_representation = ""

                for v_frame in independent_v_frames:

                    independent_v_representation += v_frame._repr_html_()
                    independent_v_representation += new_line

            if fixed_parameters_summary.empty:

                fixed_parameters_representation = empty_frame

            else:

                fixed_parameters_representation = fixed_parameters_summary._repr_html_()

        else:

            source_summary_representation = sources_summary.__repr__()

            if free_parameters_summary.empty:

                free_parameters_representation = empty_frame

            else:

                free_parameters_representation = free_parameters_summary.__repr__()

            if len(linked_frames) == 0:

                linked_summary_representation = empty_frame

            else:

                linked_summary_representation = ""

                for linked_frame in linked_frames:

                    linked_summary_representation += linked_frame.__repr__()
                    linked_summary_representation += "%s%s" % (new_line, new_line)

            if len(independent_v_frames) == 0:

                independent_v_representation = empty_frame

            else:

                independent_v_representation = ""

                for v_frame in independent_v_frames:

                    independent_v_representation += v_frame.__repr__()
                    independent_v_representation += "%s%s" % (new_line, new_line)

            if fixed_parameters_summary.empty:

                fixed_parameters_representation = empty_frame

            else:

                fixed_parameters_representation = fixed_parameters_summary.__repr__()

        # Build the representation

        representation = "Model summary:%s" % (new_line)

        if not rich_output:

            representation += "==============%s%s" % (new_line, new_line)

        else:

            representation += new_line

        # Summary on sources

        representation += source_summary_representation

        representation += new_line

        # Free parameters

        representation += "%sFree parameters (%i):%s" % (
            new_line,
            len(free_parameters),
            new_line,
        )

        if not rich_output:

            representation += "--------------------%s%s" % (new_line, new_line)

        else:

            representation += new_line

        representation += free_parameters_representation

        representation += new_line

        # Fixed parameters

        n_fix = len(parameters) - len(free_parameters) - len(linked_parameters)

        representation += "%sFixed parameters (%i):%s" % (new_line, n_fix, new_line)

        if self._complete_display:

            if not rich_output:

                representation += "---------------------%s%s" % (new_line, new_line)

            else:

                representation += new_line

            representation += fixed_parameters_representation

        else:

            representation += (
                "(abridged. Use complete=True to see all fixed parameters)%s" % new_line
            )

        representation += new_line

        # Linked parameters

        representation += "%sLinked parameters (%i):%s" % (
            new_line,
            len(self.linked_parameters),
            new_line,
        )

        if not rich_output:

            representation += "----------------------%s%s" % (new_line, new_line)

        else:

            representation += new_line

        representation += linked_summary_representation

        # Independent variables

        representation += "%sIndependent variables:%s" % (new_line, new_line)

        if not rich_output:

            representation += "----------------------%s%s" % (new_line, new_line)

        else:

            representation += new_line

        representation += independent_v_representation

        return representation

    def to_dict_with_types(self):

        # Get the serialization dictionary

        data = self.to_dict()

        # Add the types to the sources

        for key in list(data.keys()):

            try:

                element = self._get_child(key)

            except KeyError:  # pragma: no cover

                raise RuntimeError("Source %s is unknown" % key)

            else:

                # There are three possible cases. Either the element is a source, or it is an independent
                # variable, or a parameter

                if hasattr(element, "source_type"):

                    # Change the name of the key adding the source type

                    data["%s (%s)" % (key, element.source_type)] = data.pop(key)

                elif isinstance(element, IndependentVariable):

                    data["%s (%s)" % (key, "IndependentVariable")] = data.pop(key)

                elif isinstance(element, Parameter):

                    data["%s (%s)" % (key, "Parameter")] = data.pop(key)

                else:  # pragma: no cover

                    raise ModelInternalError("Found an unknown class at the top level")

        return data

    def save(self, output_file, overwrite=False):

        """Save the model to disk"""

        if os.path.exists(output_file) and overwrite is False:

            raise ModelFileExists(
                "The file %s exists already. If you want to overwrite it, use the 'overwrite=True' "
                "options as 'model.save(\"%s\", overwrite=True)'. "
                % (output_file, output_file)
            )

        else:

            data = self.to_dict_with_types()

            # Write it to disk

            try:

                # Get the YAML representation of the data

                representation = my_yaml.dump(data, default_flow_style=False)

                with open(output_file, "w+") as f:

                    # Add a new line at the end of each voice (just for clarity)

                    f.write(representation.replace("\n", "\n\n"))

            except IOError:

                raise CannotWriteModel(
                    os.path.dirname(os.path.abspath(output_file)),
                    "Could not write model file %s. Check your permissions to write or the "
                    "report on the free space which follows: " % output_file,
                )

    def get_number_of_point_sources(self) ->int:
        """
        Return the number of point sources

        :return: number of point sources
        """
        return len(self._point_sources)

    def get_point_source_position(self, id) -> Tuple[float]:
        """
        Get the point source position (R.A., Dec)

        :param id: id of the source
        :return: a tuple with R.A. and Dec.
        """

        pts = list(self._point_sources.values())[id]

        return pts.position.get_ra(), pts.position.get_dec()

    def get_point_source_fluxes(self, id: int, energies: np.ndarray, tag=None) -> np.ndarray:
        """
        Get the fluxes from the id-th point source

        :param id: id of the source
        :param energies: energies at which you need the flux
        :param tag: a tuple (integration variable, a, b) specifying the integration to perform. If this
        parameter is specified then the returned value will be the average flux for the source computed as the integral
        between a and b over the integration variable divided by (b-a). The integration variable must be an independent
        variable contained in the model. If b is None, then instead of integrating the integration variable will be
        set to a and the model evaluated in a.
        :return: fluxes
        """
        
        return list(self._point_sources.values())[id](energies, tag=tag)

    def get_point_source_name(self, id: int) -> str:

        return list(self._point_sources.values())[id].name

    def get_number_of_extended_sources(self) -> int:
        """
        Return the number of extended sources

        :return: number of extended sources
        """
        return len(self._extended_sources)

    def get_extended_source_fluxes(self, id: int, j2000_ra: float, j2000_dec: float, energies: np.ndarray) -> np.ndarray:
        """
        Get the flux of the id-th extended sources at the given position at the given energies

        :param id: id of the source
        :param j2000_ra: R.A. where the flux is desired
        :param j2000_dec: Dec. where the flux is desired
        :param energies: energies at which the flux is desired
        :return: flux array
        """

        return list(self._extended_sources.values())[id](j2000_ra, j2000_dec, energies)

    def get_extended_source_name(self, id: int) -> str:
        """
        Return the name of the n-th extended source

        :param id: id of the source (integer)
        :return: the name of the id-th source
        """

        return list(self._extended_sources.values())[id].name

    def get_extended_source_boundaries(self, id: int):

        (ra_min, ra_max), (dec_min, dec_max) = list(self._extended_sources.values())[
            id
        ].get_boundaries()

        return ra_min, ra_max, dec_min, dec_max

    def is_inside_any_extended_source(self, j2000_ra: float, j2000_dec: float) -> bool:

        for ext_source in list(self.extended_sources.values()):

            (ra_min, ra_max), (dec_min, dec_max) = ext_source.get_boundaries()

            # Transform from the 0...360 convention to the -180..180 convention, so that
            # the comparison is easier
            if ra_min > 180:

                ra_min = -(360 - ra_min)

            if ra_min <= j2000_ra <= ra_max and dec_min <= j2000_dec <= dec_max:

                return True

        # If we are here, it means that no extended source contains the provided coordinates

        return False

    def get_number_of_particle_sources(self) -> int:
        """
        Return the number of particle sources

        :return: number of particle sources
        """

        return len(self._particle_sources)

    def get_particle_source_fluxes(self, id: int, energies: np.ndarray) -> np.ndarray:
        """
        Get the fluxes from the id-th point source

        :param id: id of the source
        :param energies: energies at which you need the flux
        :return: fluxes
        """

        return list(self._particle_sources.values())[id](energies)

    def get_particle_source_name(self, id: int) -> str:

        return list(self._particle_sources.values())[id].name

    def get_total_flux(self, energies: np.ndarray) -> float:
        """
        Returns the total differential flux at the provided energies from all *point* sources

        :return:
        """

        fluxes = []

        for src in self._point_sources:

            fluxes.append(self._point_sources[src](energies))

        return np.sum(fluxes, axis=0)



import math
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional

import astropy.units as astropy_units
import numba as nb
import numpy as np
from astropy.io import fits
from interpolation import interp

from astromodels.functions.function import Function1D, FunctionMeta
from astromodels.utils import _get_data_file_path
from astromodels.utils.configuration import astromodels_config
from astromodels.utils.logging import setup_logger

log = setup_logger(__name__)

try:

    # ebltable is a Python packages to read in and interpolate tables for the photon density of
    # the Extragalactic Background Light (EBL) and the resulting opacity for high energy gamma
    # rays.

    import ebltable.tau_from_model as ebltau

    has_ebltable = True

except ImportError:

    if astromodels_config.logging.startup_warnings:

        log.warning(
            "The ebltable package is not available. Models that depend on it will not be available"
        )

    has_ebltable = False


class EBLTableNotAvailable(ImportWarning):
    pass


class InvalidUsageForFunction(Exception):
    pass


@dataclass(frozen=False)
class AbundanceTable:
    name: str
    tables: List[str]
    _current_table: str

    def set_table(self, table: str):

        """
        set the current table from AG89, WILM or ASPL

        :param table: 
        :type table: str
        :returns: 

        """
        
        
        
        old_table = self._current_table

        self._current_table = table
        
        
        if self.current_table not in self.tables:

            log.error(f"{self.name} does not contain {table} choose {','.join(self.table)}")

            self._current_table = old_table
            
            raise AssertionError()
            

    @property
    def current_table(self) -> str:

        convert= {
        "AG89": "angr",
        "ASPL": "aspl",
        "WILM": "wilm"
        }

        return convert[self._current_table]
        
    @property
    def info(self) -> str:

        _abund_info = {}
        _abund_info["WILM"] = "wilms\nfrom Wilms, Allen & McCray (2000), ApJ 542, 914 \n except for elements not listed which are given zero abundance)\n https://heasarc.nasa.gov/xanadu/xspec/manual/XSabund.html "
        _abund_info["AG89"] = "angr\nfrom Anders E. & Grevesse N. (1989, Geochimica et Cosmochimica Acta 53, 197)\n https://heasarc.nasa.gov/xanadu/xspec/manual/XSabund.html"
        _abund_info["ASPL"] = "aspl\nfrom Asplund M., Grevesse N., Sauval A.J. & Scott P. (2009, ARAA, 47, 481)\nhttps://heasarc.nasa.gov/xanadu/xspec/manual/XSabund.html"

        return _abund_info[self._current_table]

    @property
    def xsect_table(self) -> np.ndarray:

        """
        returns the XSECT table for the current model

        :returns: 

        """
        _path: Path = Path(
        "xsect") / f"xsect_{self.name}_{self.current_table}.fits"

        path_to_xsect: Path = _get_data_file_path(_path)

        with fits.open(path_to_xsect) as fxs:

            dxs = fxs[1].data
            xsect_ene = dxs["ENERGY"]
            xsect_val = dxs["SIGMA"]

        return np.array(xsect_ene, dtype=np.float64), np.array(xsect_val,
                                                           dtype=np.float64)


        
phabs = AbundanceTable("phabs", ["angr","aspl"], "AG89")
tbabs = AbundanceTable("tbabs", ["angr","aspl","wilm"], "WILM")
wabs = AbundanceTable("wabs", ["angr"], "AG89")
    


# PhAbs class


class PhAbs(Function1D, metaclass=FunctionMeta):
    r"""
    description :
        Photometric absorption (phabs implementation), f(E) = exp(- NH * sigma(E))
        contributed by Dominique Eckert
    parameters :
        NH :
            desc : absorbing column density in units of 1e22 particles per cm^2
            initial value : 1.0
            is_normalization : False
            transformation : log10
            min : 1e-4
            max : 1e4
            delta : 0.1

        redshift :
            desc : the redshift of the source
            initial value : 0.
            is_normalization : False
            min : 0
            max : 15
            delta : 0.1
            fix: True

    """
    def _setup(self):
        self._fixed_units = (astropy_units.keV,
                             astropy_units.dimensionless_unscaled)
        self.init_xsect(astromodels_config.absorption_models.phabs_table.value)

    def _set_units(self, x_unit, y_unit):
        self.NH.unit = astropy_units.cm**(-2)
        self.redshift.unit = astropy_units.dimensionless_unscaled

    def init_xsect(self, abund_table="AG89"):
        """
        Set the abundance table

        :param abund_table: "ASPL", "AG89" 
        :returns: 
        :rtype: 

        """

        # load cross section data

        try:
            phabs.set_table(abund_table)

            log.debug(f"phabs model set to {abund_table}")
            
        except:

            log.info(f"defaulting to {astromodels_config.absorption_models.phabs_table.value}")

            phabs.set_table(astromodels_config.absorption_models.phabs_table.value)
            

        self.xsect_ene, self.xsect_val = phabs.xsect_table

    @property
    def abundance_table(self):
        print(phabs.info)

        
    def evaluate(self, x, NH, redshift):

        if isinstance(x, astropy_units.Quantity):

            _unit = astropy_units.cm**2
            _y_unit = astropy_units.dimensionless_unscaled
            _x = x.value
            _redshift = redshift.value
        else:

            _unit = 1.0
            _y_unit = 1.0
            _redshift = redshift
            _x = x

        xsect_interp = interp(self.xsect_ene, self.xsect_val,
                              _x * (1 + _redshift))

        # evaluate the exponential with numba
        
        spec = _numba_eval(NH, xsect_interp) * _y_unit
        
        return spec


# TbAbs class


class TbAbs(Function1D, metaclass=FunctionMeta):
    r"""
    description :
        Photometric absorption (Tbabs implementation), f(E) = exp(- NH * sigma(E))
        contributed by Dominique Eckert
    parameters :
        NH :
            desc : absorbing column density in units of 1e22 particles per cm^2
            initial value : 1.0
            is_normalization : True
            transformation : log10
            min : 1e-4
            max : 1e4
            delta : 0.1

        redshift :
            desc : the redshift of the source
            initial value : 0.
            is_normalization : False
            min : 0
            max : 15
            delta : 0.1
            fix: True

    """
    def _setup(self):

        self.init_xsect(astromodels_config.absorption_models.tbabs_table.value)

        self._fixed_units = (astropy_units.keV,
                             astropy_units.dimensionless_unscaled)

    def _set_units(self, x_unit, y_unit):

        self.NH.unit = astropy_units.cm**(-2)
        self.redshift.unit = astropy_units.dimensionless_unscaled

    def init_xsect(self, abund_table="WILM"):
        """
        Set the abundance table

        :param abund_table: "WILM", "ASPL", "AG89" 
        :returns: 
        :rtype: 

        """
        
        try:
            tbabs.set_table(abund_table)
            
            log.debug(f"tbabs model set to {abund_table}")
            
        except:

            log.info(f"defaulting to {astromodels_config.absorption_models.tbabs_table.value}")

            tbabs.set_table(astromodels_config.absorption_models.tbabs_table.value)
            

        self.xsect_ene, self.xsect_val = tbabs.xsect_table

    @property
    def abundance_table(self):
        print(tbabs.info)

    def evaluate(self, x, NH, redshift):

        if isinstance(x, astropy_units.Quantity):

            _unit = astropy_units.cm**2
            _y_unit = astropy_units.dimensionless_unscaled
            _x = x.value
            _redshift = redshift.value

        else:

            _unit = 1.0
            _y_unit = 1.0
            _redshift = redshift
            _x = x

        xsect_interp = interp(self.xsect_ene, self.xsect_val,
                              _x * (1 + _redshift))

        spec = _numba_eval(NH, xsect_interp) * _y_unit

        return spec


# WAbs class


class WAbs(Function1D, metaclass=FunctionMeta):
    r"""
    description :
        Photometric absorption (Wabs implementation), f(E) = exp(- NH * sigma(E))
        contributed by Dominique Eckert
    parameters :
        NH :
            desc : absorbing column density in units of 1e22 particles per cm^2
            initial value : 1.0
            is_normalization : True
            transformation : log10
            min : 1e-4
            max : 1e4
            delta : 0.1
        redshift :
            desc : the redshift of the source
            initial value : 0.
            is_normalization : False
            min : 0
            max : 15
            delta : 0.1
            fix: True

    """

    def _setup(self):
        self._fixed_units = (astropy_units.keV,
                             astropy_units.dimensionless_unscaled)
        self.init_xsect()

    def _set_units(self, x_unit, y_unit):
        self.NH.unit = astropy_units.cm**(-2)
        self.redshift.unit = astropy_units.dimensionless_unscaled

    def init_xsect(self):
        """
        Set the abundance table

        :returns:
        :rtype:

        """            

        self.xsect_ene, self.xsect_val = wabs.xsect_table


    @property
    def abundance_table(self):
        print(wabs.info)


    def evaluate(self, x, NH, redshift):

        if isinstance(x, astropy_units.Quantity):

            _unit = astropy_units.cm**2
            _y_unit = astropy_units.dimensionless_unscaled
            _x = x.value
            _redshift = redshift.value
        else:

            _unit = 1.0
            _y_unit = 1.0
            _redshift = redshift
            _x = x

        xsect_interp = interp(self.xsect_ene, self.xsect_val,
                              _x * (1 + _redshift))

        xsect_interp = interp( self.xsect_ene, self.xsect_val, _x * (1+ _redshift))

        spec = _numba_eval(NH,xsect_interp) * _y_unit

        return spec



if has_ebltable:

    class EBLattenuation(Function1D, metaclass=FunctionMeta):
        r"""
        description :
            Attenuation factor for absorption in the extragalactic background light (EBL) ,
            to be used for extragalactic source spectra. Based on package "ebltable" by
            Manuel Meyer, https://github.com/me-manu/ebltable .

        latex: not available

        parameters :

          redshift :
                desc : redshift of the source
                initial value : 1.0
                fix : yes

          attenuation :
                desc : scaling factor for the strength of attenuation
                initial value : 1.0
                min : 0.0
                max : 10.0
                fix : yes

        """

        def _setup(self):

            # define EBL model, use dominguez as default
            self._tau = ebltau.OptDepth.readmodel(model=astromodels_config.absorption_models.ebl_table.value)

        def set_ebl_model(self, modelname):

            # passing modelname to ebltable, which will check if defined
            self._tau = ebltau.OptDepth.readmodel(model=modelname)

        def _set_units(self, x_unit, y_unit):

            if not hasattr(x_unit, "physical_type") or x_unit.physical_type != "energy":

                # x should be energy
                raise InvalidUsageForFunction(
                    "Unit for x is not an energy. The function "
                    "EBLOptDepth calculates energy-dependent "
                    "absorption."
                )

            # y should be dimensionless
            if (
                not hasattr(y_unit, "physical_type")
                or y_unit.physical_type != "dimensionless"
            ):
                raise InvalidUsageForFunction(
                    "Unit for y is not dimensionless.")

            self.redshift.unit = astropy_units.dimensionless_unscaled
            self.attenuation.unit = astropy_units.dimensionless_unscaled

        def evaluate(self, x, redshift, attenuation):

            if isinstance(x, astropy_units.Quantity):

                # ebltable expects TeV
                eTeV = x.to(astropy_units.TeV).value
                _unit = astropy_units.dimensionless_unscaled
                _redshift = redshift.value
                _attenuation = attenuation.value

            else:

                # otherwise it's in keV
                eTeV = x/1.e9

                _unit = 1.
                _redshift = redshift
                _attenuation = attenuation
                
                
            return _numba_eval(self._tau.opt_depth(_redshift, eTeV), _attenuation) * _unit



@nb.vectorize
def _exp(x):
    return math.exp(x)

@nb.njit(fastmath=True)
def _numba_eval(nh, xsect_interp):

    return _exp(-nh * xsect_interp )




__author__ = 'giacomov'

from enum import Enum, unique
from astromodels.utils.logging import setup_logger
from typing import Dict, Optional, List, Any

from astromodels.core.parameter import Parameter

import collections

log = setup_logger(__name__)

@unique
class SourceType(Enum):
    PARTICLE_SOURCE = 'particle source'
    POINT_SOURCE = 'point source'
    EXTENDED_SOURCE = 'extended source'

    def __str__(self):
        return f"{self.value}"


class UnknownSourceType(Exception):
    pass


class Source(object):

    def __init__(self, list_of_components: List[Any], src_type: str, spatial_shape=None):

        # Make the dictionary of components
        self._components: Dict[str, Any] = collections.OrderedDict()

        for component in list_of_components:

            self._components[component.name] = component

        if src_type not in SourceType:

            log.error(f"Source of type {src_type} is unknown")
            
            raise UnknownSourceType()

        else:

            # Store the type string
            self._src_type = src_type

    def has_free_parameters(self):

        raise NotImplementedError("You need to override this")

    @property
    def free_parameters(self) -> Dict[str, Parameter]:
        """
        Returns a dictionary of free parameters for this source

        :return:
        """

        raise NotImplementedError("You need to override this")

    @property
    def components(self) -> Dict[str, Any]:
        """
        Return the dictionary of components

        :return: dictionary of components
        """

        return self._components

    @property
    def source_type(self) -> str:
        """
        Return the type of the source ('point source' or 'extended source')

        :return: type of the source
        """

        return self._src_type



from __future__ import print_function
from __future__ import division
import pytest
import astropy.units as u
from astromodels import clone_model, PointSource, Model, load_model
from pathlib import Path
try:

    from astromodels.xspec import *

except:

    has_XSPEC = False

else:

    has_XSPEC = True


# This defines a decorator which can be applied to single tests to
# skip them if the condition is not met
skip_if_xspec_is_not_available = pytest.mark.skipif(not has_XSPEC,
                                                    reason="XSPEC not available")


@skip_if_xspec_is_not_available
def test_xspec_load():

    # no need to do anything really
    s = XS_phabs() * XS_powerlaw() + XS_bbody()
    print(s(1.0))
    s.set_units(u.keV, 1 / (u.keV * u.cm**2 * u.s))
    print(s(1.0 * u.keV)) 


@skip_if_xspec_is_not_available
def test_xspec_saving():


    s =  XS_powerlaw() + XS_bbody()

    ps = PointSource("test", 0, 0, spectral_shape=s)

    model = Model(ps)


    cloned_model = clone_model(model)

    filename = "_test_xspec_model.yml"
    
    model.save(filename)

    reloaded_model = load_model(filename)


    p = Path(filename)

    p.unlink()



# NOTE: the astropy.units.format pre-defined formatters use ply.yacc
# for parsing, which unfortunately is not thread safe. This causes
# all sort of weird problems when trying to use ipyparallel features.
# Here we implement a format which is very simple, does not use
# ply.yacc and is thread safe

from builtins import str
from builtins import map
from builtins import zip
import re
from astropy.units.format.base import Base
import astropy.units as u
from functools import reduce


# NOTE: the metaclass in Base will take care of registering
# this format, which will be available in the u.Unit
# constructor and in the .to_string method of the Unit and
# Quantity classes as format='threadsafe' (case insensitive)

def _format_one(xxx_todo_changeme):
    # This is for example 'cm-2' if base=cm and power=-2,
    # but only 'cm' if base=cm and power=1

    (base, power) = xxx_todo_changeme
    return "%s%s" % (base.to_string(), power if power != 1 else '')


class ThreadSafe(Base):

    @classmethod
    def parse(cls, s):

        # Since often we get these strings from a YAML file,
        # we want to make sure there is no new line within
        # the string.
        assert "\n" not in s

        # We assume that the string is something like
        # [unit1]{power} [unit2]{power} ..., like for example:
        # "m s-1" (a speed) or "keV-1 cm-2 s-1"
        # (a differential flux in 1 / (keV cm**2 s))
        # This is of course the format that is the output of
        # our to_string method(). See there for details

        tokens = re.findall('([a-zA-z]+)(-?\+?[0-9]+)?', s)

        # tokens is a list of tuples of the type [(unit name, power), ...]
        # Here we build a list like [u.m, u.s**(-1), ...]
        r = []

        for (unit, power) in tokens:

            # Get the primitive unit from the units module
            thisr = getattr(u, unit)

            # If needed, raise it to the power
            if power != '':

                thisr = thisr**power

            r.append(thisr)

        # Now we multiply the units
        # so that we get something like u.m * u.s**(-1)

        return reduce(lambda x, y: x*y, r)

    @classmethod
    def to_string(cls, unit):

        # The Unit class has two lists: bases (which are primitive units)
        # and powers (the correspondent powers). Thus we just need to use
        # those and encode them in a string like [unit1]{power} [unit2]{power} ...

        tokens = list(map(_format_one, list(zip(unit.bases, unit.powers))))

        return str(" ".join(tokens))



from __future__ import division
from past.utils import old_div
__author__ = 'giacomov'

import collections

import numpy

from astromodels.core.spectral_component import SpectralComponent
from astromodels.core.tree import Node
from astromodels.core.units import get_units
from astromodels.sources.source import Source, SourceType
from astromodels.utils.pretty_list import dict_to_list
from astromodels.utils.logging import setup_logger


log = setup_logger(__name__)

class ParticleSource(Source, Node):
    """
    A source of particles with a certain distribution. This source does not produce any electromagnetic signal,
    but it is useful if used in conjuction with spectral shapes which need a particle distribution as input, like
    for example a Synchrotron kernel.

    :param name: name for the source
    :param distribution_shape: a function describing the energy distribution of the particles
    :param components: a list of SpectralComponents instances
    :return:

    """

    def __init__(self, name, distribution_shape=None, components=None):

        Node.__init__(self, name)

        if components is None:

            if distribution_shape is None:

                log.error("You have to either provied a list of components, or a " \
                                                   "distribution shape")

                raise AssertionError()

            components = [SpectralComponent("main", distribution_shape)]

        Source.__init__(self, components, SourceType.PARTICLE_SOURCE)

        # Add a node called 'spectrum'

        spectrum_node = Node('spectrum')
        spectrum_node._add_children(list(self._components.values()))

        self._add_child(spectrum_node)

        type(self).__call__ = type(self).get_flux

        # Set the units
        # Now sets the units of the parameters for the energy domain

        current_units = get_units()

        # energy as x and particle flux as y
        x_unit = current_units.energy
        y_unit = old_div(1, current_units.energy)

        # Now set the units of the components
        for component in list(self._components.values()):
            component.shape.set_units(x_unit, y_unit)

    def get_flux(self, energies):

        """Get the total flux of this particle source at the given energies (summed over the components)"""

        results = [component.shape(energies) for component in list(self.components.values())]

        return numpy.sum(results, 0)

    def _repr__base(self, rich_output=False):
        """
        Representation of the object

        :param rich_output: if True, generates HTML, otherwise text
        :return: the representation
        """

        # Make a dictionary which will then be transformed in a list

        repr_dict = collections.OrderedDict()

        key = '%s (particle source)' % self.name

        repr_dict[key] = collections.OrderedDict()
        repr_dict[key]['spectrum'] = collections.OrderedDict()

        for component_name, component in list(self.components.items()):

            repr_dict[key]['spectrum'][component_name] = component.to_dict(minimal=True)

        return dict_to_list(repr_dict, rich_output)

    @property
    def free_parameters(self):
        """
        Returns a dictionary of free parameters for this source.
        We use the parameter path as the key because it's 
        guaranteed to be unique, unlike the parameter name.

        :return:
        """
        free_parameters = collections.OrderedDict()

        for component in self._components.values():

            for par in component.shape.parameters.values():

                if par.free:

                    free_parameters[par.path] = par

        return free_parameters

    @property
    def parameters(self):
        """
        Returns a dictionary of all parameters for this source.
        We use the parameter path as the key because it's 
        guaranteed to be unique, unlike the parameter name.

        :return:
        """
        all_parameters = collections.OrderedDict()

        for component in self._components.values():

            for par in component.shape.parameters.values():

                all_parameters[par.path] = par

        return all_parameters





from __future__ import print_function

import astropy.units as astropy_units
import numpy as np
import six
from scipy.interpolate import RegularGridInterpolator

from astromodels.functions.function import Function1D, FunctionMeta
from astromodels.utils import _get_data_file_path
from astromodels.utils.logging import setup_logger

log = setup_logger(__name__)


class DMFitFunction(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        Class that evaluates the spectrum for a DM particle of a given
        mass, channel, cross section, and J-factor. Based on standard
        Fermi Science Tools function DMFitFunction. Note input table only
        calculated spectra up to m_DM of 10 TeV

        The parameterization is given by

        F(x) = 1 / (8 * pi) * (1/mass^2) * sigmav * J * dN/dE(E,mass,i)

    latex : $$

    parameters :

        mass :
            desc : DM mass (GeV)
            initial value : 10
            fix : yes

        channel :
            desc : DM annihilation channel
            initial value : 4
            fix : yes

        sigmav :
            desc : DM annihilation cross section (cm^3/s)
            initial value : 1.e-26

        J :
            desc : Target total J-factor (GeV^2 cm^-5)
            initial value : 1.e20
            fix : yes
    """

    def _setup(self):

        tablepath = _get_data_file_path("dark_matter/gammamc_dif.dat")

        self._data = np.loadtxt(tablepath)

        """
            Mapping between the channel codes and the rows in the gammamc file

            1 : 8, # ee
            2 : 6, # mumu
            3 : 3, # tautau
            4 : 1, # bb
            5 : 2, # tt
            6 : 7, # gg
            7 : 4, # ww
            8 : 5, # zz
            9 : 0, # cc
            10 : 10, # uu
            11 : 11, # dd
            12 : 9, # ss
        """

        channel_index_mapping = {
            1: 8,  # ee
            2: 6,  # mumu
            3: 3,  # tautau
            4: 1,  # bb
            5: 2,  # tt
            6: 7,  # gg
            7: 4,  # ww
            8: 5,  # zz
            9: 0,  # cc
            10: 10,  # uu
            11: 11,  # dd
            12: 9,  # ss
        }

        # Number of decades in x = log10(E/M)
        ndec = 10.0
        xedge = np.linspace(0, 1.0, 251)
        self._x = 0.5 * (xedge[1:] + xedge[:-1]) * ndec - ndec

        ichan = channel_index_mapping[int(self.channel.value)]

        # These are the mass points
        self._mass = np.array(
            [
                2.0,
                4.0,
                6.0,
                8.0,
                10.0,
                25.0,
                50.0,
                80.3,
                91.2,
                100.0,
                150.0,
                176.0,
                200.0,
                250.0,
                350.0,
                500.0,
                750.0,
                1000.0,
                1500.0,
                2000.0,
                3000.0,
                5000.0,
                7000.0,
                1e4,
            ]
        )
        self._dn = self._data.reshape((12, 24, 250))

        self._dn_interp = RegularGridInterpolator(
            [self._mass, self._x],
            self._dn[ichan, :, :],
            bounds_error=False,
            fill_value=None,
        )

        if self.mass.value > 10000:

            print("Warning: DMFitFunction only appropriate for masses <= 10 TeV")
            print("To model DM from 2 GeV < mass < 1 PeV use DMSpectra")

    def _set_units(self, x_unit, y_unit):

        # Usually a model should not assume fixed units for energy or anything else. However,
        # in this case this model is so specialistic that we can assume GeV

        self.mass.unit = astropy_units.GeV

        self.channel.unit = astropy_units.dimensionless_unscaled

        self.sigmav.unit = astropy_units.cm ** 3 / astropy_units.s

        self.J.unit = astropy_units.GeV ** 2 / astropy_units.cm ** 5

    def print_channel_mapping(self):

        channel_mapping = {
            1: "ee",
            2: "mumu",
            3: "tautau",
            4: "bb",
            5: "tt",
            6: "gg",
            7: "ww",
            8: "zz",
            9: "cc",
            10: "uu",
            11: "dd",
            12: "ss",
        }

        print(channel_mapping)

        return channel_mapping

    # noinspection PyPep8Naming
    def evaluate(self, x, mass, channel, sigmav, J):

        if isinstance(x, astropy_units.Quantity):

            # We need to convert to GeV
            xx = x.to(astropy_units.GeV)

        else:

            # We can assume that the input is in keV

            keVtoGeV = 1e-6

            # xm expects gamma ray energies in MeV
            xx = np.multiply(x, keVtoGeV)

        xm = np.log10(np.divide(xx, mass))
        phip = (
            1.0 / (8.0 * np.pi) * np.power(mass, -2) * (sigmav * J)
        )  # units of this should be 1 / cm**2 / s
        dn = self._dn_interp((mass, xm))
        dn[xm > 0] = 0

        return np.multiply(phip, np.divide(dn, x))


class DMSpectra(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        Class that evaluates the spectrum for a DM particle of a given
        mass, channel, cross section, and J-factor. Combines Pythia-based tables
        from both Fermi (2 GeV < m_DM < 10 TeV) and HAWC (10 TeV < m_dm < 1 PeV)

        The parameterization is given by

        F(x) = 1 / (8 * pi) * (1/mass^2) * sigmav * J * dN/dE(E,mass,i)

        Note that this class assumes that mass and J-factor are provided
        in units of GeV and GeV^2 cm^-5

    latex : $$

    parameters :

        mass :
            desc : DM mass (GeV)
            initial value : 10
            fix : yes

        channel :
            desc : DM annihilation channel
            initial value : 4
            fix : yes

        sigmav :
            desc : DM annihilation cross section (cm^3/s)
            initial value : 1.e-26

        J :
            desc : Target total J-factor (GeV^2 cm^-5)
            initial value : 1.e20
            fix : yes
    """

    def _setup(self):

        # Get and open the two data files

        tablepath_h = _get_data_file_path("dark_matter/dmSpecTab.npy")
        self._data_h = np.load(tablepath_h)

        tablepath_f = _get_data_file_path("dark_matter/gammamc_dif.dat")
        self._data_f = np.loadtxt(tablepath_f)

        """
            Mapping between the channel codes and the rows in the gammamc file
            dmSpecTab.npy created to match this mapping too

            1 : 8, # ee
            2 : 6, # mumu
            3 : 3, # tautau
            4 : 1, # bb
            5 : 2, # tt
            6 : 7, # gg
            7 : 4, # ww
            8 : 5, # zz
            9 : 0, # cc
            10 : 10, # uu
            11 : 11, # dd
            12 : 9, # ss
            """

        channel_index_mapping = {
            1: 8,  # ee
            2: 6,  # mumu
            3: 3,  # tautau
            4: 1,  # bb
            5: 2,  # tt
            6: 7,  # gg
            7: 4,  # ww
            8: 5,  # zz
            9: 0,  # cc
            10: 10,  # uu
            11: 11,  # dd
            12: 9,  # ss
        }

        # Number of decades in x = log10(E/M)
        ndec = 10.0
        xedge = np.linspace(0, 1.0, 251)
        self._x = 0.5 * (xedge[1:] + xedge[:-1]) * ndec - ndec

        ichan = channel_index_mapping[int(self.channel.value)]

        # These are the mass points in GeV
        self._mass_h = np.array(
            [
                50.0,
                61.2,
                74.91,
                91.69,
                112.22,
                137.36,
                168.12,
                205.78,
                251.87,
                308.29,
                377.34,
                461.86,
                565.31,
                691.93,
                846.91,
                1036.6,
                1268.78,
                1552.97,
                1900.82,
                2326.57,
                2847.69,
                3485.53,
                4266.23,
                5221.81,
                6391.41,
                7823.0,
                9575.23,
                11719.94,
                14345.03,
                17558.1,
                21490.85,
                26304.48,
                32196.3,
                39407.79,
                48234.54,
                59038.36,
                72262.07,
                88447.7,
                108258.66,
                132506.99,
                162186.57,
                198513.95,
                242978.11,
                297401.58,
                364015.09,
                445549.04,
                545345.37,
                667494.6,
                817003.43,
                1000000.0,
            ]
        )

        # These are the mass points in GeV
        self._mass_f = np.array(
            [
                2.0,
                4.0,
                6.0,
                8.0,
                10.0,
                25.0,
                50.0,
                80.3,
                91.2,
                100.0,
                150.0,
                176.0,
                200.0,
                250.0,
                350.0,
                500.0,
                750.0,
                1000.0,
                1500.0,
                2000.0,
                3000.0,
                5000.0,
                7000.0,
                1e4,
            ]
        )

        self._mass = np.append(self._mass_f, self._mass_h[27:])

        self._dn_f = self._data_f.reshape((12, 24, 250))

        # Is this really used?
        self._dn_h = self._data_h

        self._dn = np.zeros((12, len(self._mass), 250))
        self._dn[:, 0:24, :] = self._dn_f
        self._dn[:, 24:, :] = self._dn_h[:, 27:, :]

        self._dn_interp = RegularGridInterpolator(
            [self._mass, self._x],
            self._dn[ichan, :, :],
            bounds_error=False,
            fill_value=None,
        )

        if self.channel.value in [1, 6, 7] and self.mass.value > 10000.0:
            log.error(
                "currently spectra for selected channel and mass not implemented."
            )
            log.error(
                "Spectra for channels ['ee','gg','WW'] currently not available for mass > 10 TeV"
            )

    def _set_units(self, x_unit, y_unit):

        self.mass.unit = astropy_units.GeV
        self.channel.unit = astropy_units.dimensionless_unscaled
        self.sigmav.unit = astropy_units.cm ** 3 / astropy_units.s
        self.J.unit = astropy_units.GeV ** 2 / astropy_units.cm ** 5

    def print_channel_mapping(self):
        channel_mapping = {
            1: "ee",
            2: "mumu",
            3: "tautau",
            4: "bb",
            5: "tt",
            6: "gg",
            7: "ww",
            8: "zz",
            9: "cc",
            10: "uu",
            11: "dd",
            12: "ss",
        }

        print(channel_mapping)

        return channel_mapping

    # noinspection PyPep8Naming
    def evaluate(self, x, mass, channel, sigmav, J):

        if isinstance(x, astropy_units.Quantity):

            # We need to convert to GeV
            xx = x.to(astropy_units.MeV)

        else:

            # We can assume that the input is in keV

            keVtoGeV = 1e-6

            # xm expects gamma ray energies in MeV
            xx = np.multiply(x, keVtoGeV)

        xm = np.log10(np.divide(xx, mass))

        phip = (
            1.0 / (8.0 * np.pi) * np.power(mass, -2) * (sigmav * J)
        )  # units of this should be 1 / cm**2
        dn = self._dn_interp((mass, xm))  # note this is unitless (dx = d(xm))
        dn[xm > 0] = 0

        return np.multiply(phip, np.divide(dn, x))



from __future__ import division

import collections
import gc
import os
import re
from builtins import object, range, str
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import astropy.io.fits as fits
import astropy.units as u
import h5py
import numpy as np
import scipy.interpolate
from future.utils import with_metaclass
from interpolation import interp
from interpolation.splines import eval_linear

from astromodels.core.parameter import Parameter
from astromodels.functions.function import Function1D, FunctionMeta
from astromodels.utils import get_user_data_path
from astromodels.utils.logging import setup_logger

log = setup_logger(__name__)

# A very small number which will be substituted to zero during the construction
# of the templates
_TINY_ = 1e-50

__all__ = [
    "IncompleteGrid",
    "ValuesNotInGrid",
    "MissingDataFile",
    "TemplateModelFactory",
    "TemplateModel",
]


class IncompleteGrid(RuntimeError):
    pass


class ValuesNotInGrid(ValueError):
    pass


class MissingDataFile(RuntimeError):
    pass


# This dictionary will keep track of the new classes already created in the current session
_classes_cache = {}


class GridInterpolate(object):
    def __init__(self, grid, values):
        self._grid = grid
        self._values = np.ascontiguousarray(values)

    def __call__(self, v):

        return eval_linear(self._grid, self._values, v)


class UnivariateSpline(object):
    def __init__(self, x, y):

        self._x = x
        self._y = y

    def __call__(self, v):

        return interp(self._x, self._y, v)


class TemplateModelFactory(object):
    def __init__(
        self,
        name,
        description,
        energies,
        names_of_parameters,
        interpolation_degree=1,
        spline_smoothing_factor=0,
    ):

        # Store model name

        # Enforce that it does not contain spaces nor strange characters

        name = str(name)

        if re.match("[a-zA-Z_][a-zA-Z0-9_]*", name) is None:
            log.error(
                "The provided name '%s' is not a valid name. You cannot use spaces, "
                "or special characters")
            raise RuntimeError()

        self._name = name

        self._description = str(description)

        # Store energy grid

        if not isinstance(energies, u.Quantity):

            log.warning(
                "Energy unit is not a Quantity instance, so units has not been provided. Using keV."
            )

            energies = energies * u.keV

        self._energies = np.array(energies.to(u.keV).value)

        # Enforce that they are ordered
        self._energies.sort()

        # We create a dictionary which will contain the grid for each parameter

        self._parameters_grids = collections.OrderedDict()

        for parameter_name in names_of_parameters:

            self._parameters_grids[parameter_name] = None

        self._data_frame = None

        self._interpolators = None

        self._interpolation_degree = interpolation_degree

        self._spline_smoothing_factor = int(spline_smoothing_factor)

    def define_parameter_grid(self, parameter_name: str,
                              grid: np.ndarray) -> None:
        """
        Define the parameter grid for this parameter.
        Pass the name of the parameter and the array of values that it will take in the grid
        """

        if not parameter_name in self._parameters_grids:

            log.error(f"Parameter {parameter_name} is not part of this model")

            raise AssertionError()

        grid_ = np.array(grid)

        if not (grid_.shape[0] > 1):

            log.error(
                "A grid for a parameter must contain at least two elements")

            raise AssertionError()

        # Assert that elements are unique

        if not np.all(np.unique(grid_) == grid_):

            log.error(
                f"Non-unique elements in grid for parameter {parameter_name}")

            raise AssertionError()

        self._parameters_grids[parameter_name] = grid_

    def add_interpolation_data(self, differential_fluxes,
                               **parameters_values_input):

        # Verify that the grid has been defined for all parameters

        for grid in list(self._parameters_grids.values()):

            if grid is None:

                log.error(
                    "You need to define a grid for all parameters, by using the "
                    "define_parameter_grid method.")

                raise IncompleteGrid()

        # this is the first run and we will create
        # a matrix that stores the data in an
        # n_par,..., n_energies matrix

        if self._data_frame is None:

            shape = []

            for k, v in self._parameters_grids.items():

                shape.append(len(v))

            shape.append(self._energies.shape[0])

            log.debug(f"grid shape: {shape}")

            self._data_frame = np.zeros(tuple(shape))

            log.debug(f"grid shape actual: {self._data_frame.shape}")

            # This is the first data set, create the data frame

            # Create the multi-index

            # self._multi_index = pd.MultiIndex.from_product(
            #     list(self._parameters_grids.values()),
            #     names=list(self._parameters_grids.keys()),
            # )

            # # Pre-fill the data matrix with nans, so we will know if some elements have not been filled

            # self._data_frame = pd.DataFrame(index=self._multi_index,
            #                                 columns=self._energies)

        # Make sure we have all parameters and order the values in the same way as the dictionary

        parameter_idx = []

        for i, (k, v) in enumerate(self._parameters_grids.items()):

            if not k in parameters_values_input:

                log.error(f"Parameter {k} is not in input")

                raise AssertionError()

            parameter_idx.append(
                int(np.where(v == parameters_values_input[k])[0][0]))

        log.debug(f" have index {parameter_idx}")

        # If the user did not specify one of the parameters, then the parameters_values array will contain nan

        if not len(parameter_idx) == len(self._parameters_grids):

            log.error("You didn't specify all parameters' values.")

            raise AssertionError()

        # Make sure we are dealing with pure numpy arrays (list and astropy.Quantity instances will be transformed)
        # First we transform the input into a u.Quantity (if it's not already)

        if not isinstance(differential_fluxes, u.Quantity):

            differential_fluxes = (np.array(differential_fluxes) * 1 /
                                   (u.keV * u.s * u.cm**2))  # type: u.Quantity

        # Then we transform it in the right units and we cast it back to a pure np.array

        differential_fluxes = np.array(
            differential_fluxes.to(1/ (u.keV * u.s * u.cm**2)).value)

        # Now let's check for valid inputs

        if not self._energies.shape[0] == differential_fluxes.shape[0]:

            log.error("Differential fluxes and energies must have "
                      "the same number of elements")

            raise AssertionError()

        # Check that the provided value does not contains nan, inf nor zero (as the interpolation happens in the
        # log space)
        if not np.all(np.isfinite(differential_fluxes)):

            log.error(
                "You have invalid values in the differential flux (nan or inf)"
            )

            raise AssertionError()

        if not np.all(differential_fluxes >= 0):

            log.error(
                "You have negative values in the differential flux (which is of "
                "course impossible)")

            raise AssertionError()

        if not np.all(differential_fluxes > 0):

            log.warning(
                "You have zeros in the differential flux. Since the interpolation happens in the log space, "
                "this cannot be accepted. We will substitute zeros with %g" %
                _TINY_)

            idx = differential_fluxes == 0  # type: np.ndarray
            differential_fluxes[idx] = _TINY_

        # Now set the corresponding values in the data frame

        # Now set the values in the data frame

        self._data_frame[tuple(parameter_idx)] = np.atleast_2d(
            differential_fluxes)

    def save_data(self, overwrite: bool = False):

        # First make sure that the whole data matrix has been filled

        if np.any(np.isnan(self._data_frame)):

            log.error("You have NaNs in the data matrix. Usually this means "
                      "that you didn't fill it up completely, or that some of "
                      "your data contains nans. Cannot save the file.")

            raise AssertionError()

        # Get the data directory

        data_dir_path: Path = get_user_data_path()

        # Sanitize the data file

        filename_sanitized: Path = data_dir_path / f"{self._name}.h5"

        # Check that it does not exists

        if filename_sanitized.exists():

            if overwrite:

                try:

                    os.remove(filename_sanitized)

                except:

                    log.error(
                        "The file %s already exists and cannot be removed (maybe you do not have "
                        "permissions to do so?). " % filename_sanitized)

                    raise IOError()

            else:

                log.error(
                    "The file %s already exists! You cannot call two different "
                    "template models with the same name" % filename_sanitized)

                raise IOError()

        # Open the HDF5 file and write objects

        template_file: TemplateFile = TemplateFile(
            name=self._name,
            description=self._description,
            spline_smoothing_factor=self._spline_smoothing_factor,
            interpolation_degree=self._interpolation_degree,
            grid=self._data_frame,
            energies=self._energies,
            parameters=self._parameters_grids,
            parameter_order=list(self._parameters_grids.keys()))

        template_file.save(filename_sanitized)

   

# This adds a method to a class at runtime


def add_method(self, method, name=None):

    if name is None:

        name = method.__name__

    setattr(self.__class__, name, method)


class RectBivariateSplineWrapper(object):
    """
    Wrapper around RectBivariateSpline, which supplies a __call__ method which accept the same
    syntax as the other interpolation methods

    """
    def __init__(self, *args, **kwargs):

        # We can use interp2, which features spline interpolation instead of linear interpolation

        self._interpolator = scipy.interpolate.RectBivariateSpline(
            *args, **kwargs)

    def __call__(self, x):

        res = self._interpolator(*x)

        return res[0][0]


@dataclass
class TemplateFile:
    """
    simple container to read and write 
    the data to an hdf5 file

    """
    name: str
    description: str
    grid: np.ndarray
    parameters: Dict[str, np.ndarray]
    parameter_order: List[str]
    energies: np.ndarray
    interpolation_degree: int
    spline_smoothing_factor: float


    def save(self, file_name: str):

        """
        serialize the contents to a file

        :param file_name: 
        :type file_name: str
        :returns: 

        """
        with h5py.File(file_name, "w") as f:

            f.attrs["name"] = self.name
            f.attrs["description"] = self.description
            f.attrs["interpolation_degree"] = self.interpolation_degree
            f.attrs["spline_smoothing_factor"] = self.spline_smoothing_factor

            f.create_dataset("energies", data=self.energies, compression="gzip")

            f.create_dataset("grid", data=self.grid, compression="gzip")

            # store the parameter order
            dt = h5py.special_dtype(vlen=str)
            po = np.array(self.parameter_order, dtype=dt) 
            f.create_dataset('parameter_order', data=po)
            par_group = f.create_group("parameters")
            for k in self.parameter_order:

                par_group.create_dataset(k, data=self.parameters[k], compression="gzip")

    @classmethod
    def from_file(cls, file_name: str):

        """
        read contents from a file

        :param cls: 
        :type cls: 
        :param file_name: 
        :type file_name: str
        :returns: 

        """
        
        with h5py.File(file_name, "r") as f:

            name = f.attrs["name"]
            description = f.attrs["description"] 
            interpolation_degree = f.attrs["interpolation_degree"] 
            spline_smoothing_factor = f.attrs["spline_smoothing_factor"] 

            energies = f["energies"][()]
            parameter_order = f["parameter_order"][()]

            grid = f["grid"][()]

            parameters = collections.OrderedDict()

            for k in parameter_order:

                parameters[k] = f["parameters"][k][()]
            


        return cls(name=name,
                   description=description,
                   interpolation_degree=interpolation_degree,
                   spline_smoothing_factor=spline_smoothing_factor,
                   energies=energies,
                   parameter_order=parameter_order,
                   parameters = parameters,
                   grid=grid
                   )

                
        


    
class TemplateModel(with_metaclass(FunctionMeta, Function1D)):

    r"""
    description :
        A template model
    latex : $n.a.$
    parameters :
        K :
            desc : Normalization (freeze this to 1 if the template provides the normalization by itself)
            initial value : 1.0
        scale :
            desc : Scale for the independent variable. The templates are handled as if they contains the fluxes
                   at E = scale * x.This is useful for example when the template describe energies in the rest
                   frame, at which point the scale describe the transformation between rest frame energy and
                   observer frame energy. Fix this to 1 to neutralize its effect.
            initial value : 1.0
            min : 1e-5
    """

    def _custom_init_(self, model_name, other_name=None, log_interp=True):
        """
        Custom initialization for this model

        :param model_name: the name of the model, corresponding to the root of the .h5 file in the data directory
        :param other_name: (optional) the name to be used as name of the model when used in astromodels. If None
        (default), use the same name as model_name
        :return: none
        """

        # Get the data directory

        data_dir_path: Path = get_user_data_path()

        # Sanitize the data file

        filename_sanitized: Path = data_dir_path.absolute() /  f"{model_name}.h5"
        
        if not filename_sanitized.exists():

            log.error("The data file %s does not exists. Did you use the "
                "TemplateFactory?" % (filename_sanitized))
            
            raise MissingDataFile()

        # Open the template definition and read from it

        self._data_file = filename_sanitized

        # use the file shadow to read
        
        template_file: TemplateFile = TemplateFile.from_file(filename_sanitized)

        self._parameters_grids = collections.OrderedDict()

        

        for key in template_file.parameter_order:

            try:

                # sometimes this is
                # stored binary
                
                k = key.decode()

            except(AttributeError):

                # if not, then we
                # load as a normal str
                
                k = key
                                
            log.debug(f"reading parameter {str(k)}")

  
            self._parameters_grids[str(k)] = template_file.parameters[key]

            

        self._energies = template_file.energies

        # Now get the metadata

        description = template_file.description
        name = template_file.name

        self._interpolation_degree = template_file.interpolation_degree

        self._spline_smoothing_factor = template_file.spline_smoothing_factor

        # Make the dictionary of parameters

        function_definition = collections.OrderedDict()

        function_definition["description"] = description

        function_definition["latex"] = "n.a."

        
        # Now build the parameters according to the content of the parameter grid

        parameters = collections.OrderedDict()

        parameters["K"] = Parameter("K", 1.0)
        parameters["scale"] = Parameter("scale", 1.0)

        for parameter_name in list(self._parameters_grids.keys()):

            grid = self._parameters_grids[parameter_name]

            parameters[parameter_name] = Parameter(
                parameter_name,
                np.median(grid),
                min_value=grid.min(),
                max_value=grid.max(),
            )

        if other_name is None:

            super(TemplateModel, self).__init__(name, function_definition, parameters)

        else:

            super(TemplateModel, self).__init__(
                other_name, function_definition, parameters
            )

        # Finally prepare the interpolators

        self._prepare_interpolators(log_interp, template_file.grid)

        # try can clean up the file
        
        del template_file

        gc.collect()
                
    def _prepare_interpolators(self, log_interp, data_frame):

        # Figure out the shape of the data matrices
        data_shape = [x.shape[0] for x in list(self._parameters_grids.values())]

        self._interpolators = []

        for i, energy in enumerate(self._energies):

            # Make interpolator for this energy
            # NOTE: we interpolate on the logarithm
            # unless specified

            if log_interp:

                this_data = np.array(
                    np.log10(data_frame[...,i]).reshape(*data_shape),
                    dtype=float,
                )

                self._is_log10 = True

            else:

                # work in linear space
                this_data = np.array(
                    data_frame[..., i].reshape(*data_shape), dtype=float
                )

                self._is_log10 = False

                
            if len(list(self._parameters_grids.values())) == 2:

                x, y = list(self._parameters_grids.values())

                # Make sure that the requested polynomial degree is less than the number of data sets in
                # both directions

                msg = (
                    "You cannot use an interpolation degree of %s if you don't provide at least %s points "
                    "in the %s direction. Increase the number of templates or decrease the interpolation "
                    "degree."
                )

                if len(x) <= self._interpolation_degree:

                    log.error(
                        msg
                        % (
                            self._interpolation_degree,
                            self._interpolation_degree + 1,
                            "x",
                        )
                    )

                    raise RuntimeError()

                if len(y) <= self._interpolation_degree:

                    log.error(
                        msg
                        % (
                            self._interpolation_degree,
                            self._interpolation_degree + 1,
                            "y",
                        )
                    )

                    raise RuntimeError()

                this_interpolator = RectBivariateSplineWrapper(
                    x,
                    y,
                    this_data,
                    kx=self._interpolation_degree,
                    ky=self._interpolation_degree,
                    s=self._spline_smoothing_factor,
                )

            else:

                # In more than 2d we can only use linear interpolation

                this_interpolator = GridInterpolate(
                    tuple([np.array(x) for x in list(self._parameters_grids.values())]),
                    this_data,
                )

            self._interpolators.append(this_interpolator)

        # clear the data
        self._data_frame = None
            
    def _set_units(self, x_unit, y_unit):

        self.K.unit = y_unit

        self.scale.unit = 1 / x_unit

    # This function will be substituted during construction by another version with
    # all the parameters of this template

    def evaluate(self, x, K, scale, *args):

        return K * self._interpolate(x, scale, args)

    def _interpolate(self, energies, scale, parameters_values):

        if isinstance(energies, u.Quantity):

            # Templates are always saved with energy in keV. We need to transform it to
            # a dimensionless quantity (actually we take the .value property) because otherwise
            # the logarithm below will fail.

            energies = np.array(
                energies.to("keV").value, ndmin=1, copy=False, dtype=float
            )

            # Same for the scale

            scale = scale.to(1/ u.keV).value

        if self._is_log10:

            log_energies = np.log10(energies)

        else:

            log_energies = energies

        e_tilde = self._energies * scale

        # Gather all interpolations for these parameters' values at all defined energies
        # (these are the logarithm of the values)
        # note that if these are not logged, then the name is superflous

        log_interpolations = np.array(
            [
                self._interpolators[i](np.atleast_1d(parameters_values))
                for i in range(self._energies.shape[0])
            ]
        )

        # Now interpolate the interpolations to get the flux at the requested energies

        # NOTE: the variable "interpolations" contains already the log10 of the values,

        if self._is_log10:

            interpolator = UnivariateSpline(np.log10(e_tilde), log_interpolations)

            values = np.power(10, interpolator(log_energies))

        else:

            interpolator = UnivariateSpline(e_tilde, log_interpolations)

            values = interpolator(log_energies)

        # The division by scale results from the differential:
        # E = e * scale
        # de = dE / scale
        # dN / dE = dN / de * de / dE = dN / de * (1 / scale)

        # NOTE: the units are added back through the multiplication by K in the evaluate method

        return values / scale

    def clean(self):
        """
        Table models can consume a lot of memory. If are creating lots of 
        table models in memory for simulations, you may want to call
        clean on the model try and remove some of the memory consumed by the models

        :returns: 

        """
        
        self._interpolators = None
        del self._interpolators
        gc.collect()

        log.info("You have 'cleaned' the table model at it will no longer be useable")

    def __del__(self):

        self.clean()

        
    @property
    def data_file(self):

        return self._data_file

    def to_dict(self, minimal=False):

        data = super(Function1D, self).to_dict(minimal)

        # if not minimal:
        #
        #     data['extra_setup'] = {'data_file': self._data_file}

        return data


class XSPECTableModel(object):
    def __init__(
        self,
        xspec_table_model_file,
        interpolation_degree=1,
        spline_smoothing_factor=0,
        log_centers=True,
    ):
        """
        Convert an XSPEC table model to an astromodels TemplateModel.
        usage:

        xtm = XSPECTableModel("ST95.fits")
        xtm.to_table_model('test', 'test')

        reloaded_table_model = TemplateModel('test', log_interp=False)

        Note: if the reloaded model is returning NaNs, adjust the interpolation
        scheme

        :param xspec_table_model_file:
        :param interpolation_degree:
        :param spline_smoothing_factor: spline smoothing
        :param log_centers: treat energies with log centers
        :returns:
        :rtype:

        """

        self._interpolation_degree = interpolation_degree
        self._spline_smoothing_factor = spline_smoothing_factor
        self._log_centers = log_centers

        self._xspec_file_name = xspec_table_model_file
        self._extract_model()

    def _extract_model(self):

        with fits.open(self._xspec_file_name) as f:

            # get the energies

            energies = f["ENERGIES"]
            ene_lo = energies.data["ENERG_LO"]
            ene_hi = energies.data["ENERG_HI"]

            # log centers

            if self._log_centers:
                self._energy = np.sqrt(ene_lo * ene_hi)

            else:

                self._energy = (ene_hi + ene_lo) / 2.0

            params = f["PARAMETERS"]

            self._names = params.data["NAME"]

            self._n_params = len(self._names)

            spectra = f["SPECTRA"]

            self._spectrum = spectra.data["INTPSPEC"]

            self._params_dict = {}

            for i, name in enumerate(self._names):

                this_dict = {}

                this_dict["pmin"] = params.data["MINIMUM"][i]
                this_dict["pmax"] = params.data["MAXIMUM"][i]
                if self._n_params > 1:

                    try:

                        this_dict["values"] = spectra.data["PARAMVAL{%d}" % i]

                    except:
                        this_dict["values"] = spectra.data["PARAMVAL"][:, i]

                else:

                    this_dict["values"] = spectra.data["PARAMVAL"]

                self._params_dict[name] = this_dict

    def to_table_model(self, file_name, model_name, overwrite=False):
        """
        Write the table model to your local astromodels database

        :param file_name: name of file to store
        :param model_name: name of the model
        :param overwrite: overwite the previous model
        :returns:
        :rtype:

        """

        tmf = TemplateModelFactory(
            file_name,
            model_name,
            self._energy,
            self._names,
            self._interpolation_degree,
            self._spline_smoothing_factor,
        )

        for name, param_table in self._params_dict.items():

            tmf.define_parameter_grid(name, np.unique(param_table["values"]))

        for i in range(self._spectrum.shape[0]):

            input_dict = {}
            for k, v in self._params_dict.items():
                input_dict[k] = v["values"][i]

            tmf.add_interpolation_data(self._spectrum[i, :], **input_dict)

        tmf.save_data(overwrite=overwrite)



def convert_old_table_model(model_name: str):
    from pandas import HDFStore

    # Get the data directory

    data_dir_path: Path = get_user_data_path()

    # Sanitize the data file
    
    filename_sanitized = data_dir_path / f"{model_name}.h5"
    
    
    if not filename_sanitized.exists():

        log.error("The data file %s does not exists. Did you use the "
             "TemplateFactory?" % (filename_sanitized))
         
        raise MissingDataFile( )
    
    with HDFStore(filename_sanitized) as store:
    
        data_frame = store["data_frame"]

        parameters_grids = collections.OrderedDict()

        processed_parameters = 0

        for key in list(store.keys()):

            match = re.search("p_([0-9]+)_(.+)", key)
            
            if match is None:
                
                continue
            
            else:
                
                tokens = match.groups()
                
                this_parameter_number = int(tokens[0])
                this_parameter_name = str(tokens[1])
                
                
                if not this_parameter_number == processed_parameters:

                    log.error("Parameters out of order!")

                    raise AssertionError()
                
                parameters_grids[this_parameter_name] = store[key]
                
                processed_parameters += 1
                
        energies = np.array(store["energies"])

        shape = [len(x) for _, x in parameters_grids.items()]

        shape.append(len(energies))

        grid = np.zeros(shape)

        for i, e in enumerate(energies):

            data = data_frame[e].values.reshape(*shape[:-1])

            grid[...,i] = data
        
                     
        
        # Now get the metadata
            
        metadata = store.get_storer("data_frame").attrs.metadata
        
        description = metadata["description"]
        
        name = metadata["name"]
        
        interpolation_degree = metadata["interpolation_degree"]
        
        spline_smoothing_factor = metadata["spline_smoothing_factor"]
        
        # Make the dictionary of parameters

        function_definition = collections.OrderedDict()

        function_definition["description"] = description



    template_file: TemplateFile = TemplateFile(
        name=name,
        description=description,
        spline_smoothing_factor=spline_smoothing_factor,
        interpolation_degree=interpolation_degree,
        grid=grid,
        energies=energies,
        parameters=parameters_grids,
        parameter_order=list(parameters_grids.keys()))

    template_file.save(filename_sanitized)


        



from .functions_1D import (
    Blackbody, Constant, Cubic, DiracDelta, Exponential_cutoff, Line,
    Log_parabola, Quadratic, Sin, StepFunction, StepFunctionUpper,
    has_ebltable, has_gsl, has_naima, Band, Band_Calderone, Band_grbm,
    Broken_powerlaw, Cutoff_powerlaw, Cutoff_powerlaw_Ep, Inverse_cutoff_powerlaw, Powerlaw,
    Powerlaw_Eflux, Powerlaw_flux, SmoothlyBrokenPowerLaw, Quartic,
    get_polynomial, Super_cutoff_powerlaw, PhAbs, TbAbs, WAbs, has_atomdb)

if has_naima:
    from .functions_1D import Synchrotron

if has_gsl:

    from .functions_1D import Cutoff_powerlaw_flux

if has_ebltable:
    from .functions_1D import EBLattenuation

if has_atomdb:

    from .apec import APEC, VAPEC

from .dark_matter.dm_models import DMFitFunction, DMSpectra

from .function import (Function1D, Function2D, Function3D, FunctionMeta,
                       ModelAssertionViolation)

from .functions_2D import (Asymm_Gaussian_on_sphere, Disk_on_sphere,
                           Ellipse_on_sphere, Gaussian_on_sphere,
                           Latitude_galactic_diffuse, Power_law_on_sphere,
                           SpatialTemplate_2D)
from .functions_3D import (Continuous_injection_diffusion,
                           Continuous_injection_diffusion_ellipse,
                           Continuous_injection_diffusion_legacy,
                           GalPropTemplate_3D)
from .priors import (Cauchy, Cosine_Prior, Gaussian, Log_normal,
                     Log_uniform_prior, Truncated_gaussian, Uniform_prior)
from .template_model import (MissingDataFile, TemplateModel,
                             TemplateModelFactory, XSPECTableModel)

__all__ = [
    "Band", "Band_Calderone", "Band_grbm", "Broken_powerlaw",
    "Cutoff_powerlaw", "Cutoff_powerlaw_Ep", "Inverse_cutoff_powerlaw", "Powerlaw", "Powerlaw_Eflux",
    "Powerlaw_flux", "SmoothlyBrokenPowerLaw", "Super_cutoff_powerlaw",
    "Constant", "Cubic", "DiracDelta", "Exponential_cutoff", "Line",
    "Quadratic", "Sin", "StepFunction", "StepFunctionUpper", "PhAbs", "TbAbs",
    "WAbs", "Asymm_Gaussian_on_sphere", "Disk_on_sphere", "Ellipse_on_sphere",
    "Gaussian_on_sphere", "Latitude_galactic_diffuse", "Power_law_on_sphere",
    "SpatialTemplate_2D", "Continuous_injection_diffusion",
    "Continuous_injection_diffusion_ellipse",
    "Continuous_injection_diffusion_legacy", "GalPropTemplate_3D", "DMSpectra",
    "DMFitFunction", "Cauchy", "Cosine_Prior", "Gaussian", "Log_normal",
    "Log_uniform_prior", "Truncated_gaussian", "Uniform_prior",
    "TemplateModel", "TemplateModelFactory", "XSPECTableModel",
    "MissingDataFile", "Log_parabola", "Blackbody", "Function1D", "Function2D",
    "Function3D", "FunctionMeta", "ModelAssertionViolation", "Quartic", "get_polynomial"
]

if has_atomdb:
    __all__.extend(["APEC", "VAPEC"])

if has_gsl:

    __all__.extend(["Cutoff_powerlaw_flux"])

if has_naima:

    __all__.extend(["Synchrotron"])

if has_ebltable:

    __all__.extend(["EBLattenuation"])



from builtins import object
import numpy as np


class ParameterTransformation(object):
    def __init__(self, is_positive=False):

        self._is_positive = is_positive

    @property
    def is_positive(self):
        return self._is_positive
    
    def forward(self, external_value):

        raise NotImplementedError("You have to implement this")

    def backward(self, internal_value):

        raise NotImplementedError("You have to implement this")


class LogarithmicTransformation(ParameterTransformation):

    def __init__(self):

        super(LogarithmicTransformation, self).__init__(is_positive=True)
    
    def forward(self, external_value):

        #  Throw an error if taking the logarithm of a negative number (or nan)

        with np.errstate(invalid='raise'):

            res = np.log10(external_value)

        return res

    def backward(self, internal_value):

        return 10**internal_value


_known_transformations = {'log10': LogarithmicTransformation}


def get_transformation(transformation_name):
    """
    Returns an instance of a transformation by name

    :param transformation_name:
    :return: instance of transformation with provided name
    """


    if not transformation_name in _known_transformations:

        raise ValueError("Transformation %s is not known" % transformation_name)

    else:

        return _known_transformations[transformation_name]()



__author__ = 'giacomov'

from astromodels.core.tree import Node
from astromodels.core.parameter import Parameter


class Polarization(Node):

    def __init__(self, polarization_type='linear'):

        assert polarization_type in ['linear', 'stokes'], 'polarization must be linear or stokes'

        self._polarization_type = polarization_type


        Node.__init__(self, 'polarization')


    @staticmethod
    def _get_parameter_from_input(number_or_parameter, minimum, maximum, what, desc, unit):

        # Try to transform it to float, if it works than we transform it to a parameter

        try:

            number_or_parameter = float(number_or_parameter)

        except TypeError:

            assert isinstance(number_or_parameter, Parameter), "%s must be either a number or a " \
                                                               "parameter instance" % what

            # So this is a Parameter instance already. Enforce that it has the right maximum and minimum

            parameter = number_or_parameter

            assert parameter.min_value == minimum, "%s must have a minimum of %s" % (what, minimum)
            assert parameter.max_value == maximum, "%s must have a maximum of %s" % (what, maximum)

        else:

            # This was a float. Enforce that it has a legal value

            assert minimum <= number_or_parameter <= maximum, "%s cannot have a value of %s, " \
                                                              "it must be %s <= %s <= %s" % (what, number_or_parameter,
                                                                                             minimum, what, maximum)

            parameter = Parameter(what, number_or_parameter,
                                  desc=desc, min_value=minimum, max_value=maximum, unit=unit, free=True)

        return parameter


# TODO: add transform between polarizations

class LinearPolarization(Polarization):

    def __init__(self, degree, angle):
        """
        Linear parameterization of polarization

        :param degree: The polarization degree
        :param angle: The polarization angle
        """
        super(LinearPolarization, self).__init__(polarization_type='linear')

        degree = self._get_parameter_from_input(degree, 0, 100, 'degree', 'Polarization degree', 'dimensionless_unscaled')

        angle = self._get_parameter_from_input(angle, 0, 180, 'angle', 'Polarization angle', 'deg')

        self._add_child(degree)
        self._add_child(angle)


class StokesPolarization(Polarization):

    def __init__(self, I, Q, U, V):
        """
        Stokes parameterization of polarization

        :param I:
        :param Q:
        :param U:
        :param V:
        """
        super(StokesPolarization, self).__init__(polarization_type='stokes')

        # get the parameters set up

        I = self._get_parameter_from_input(I, 0, 1, 'I', 'Stokes I')
        Q = self._get_parameter_from_input(Q, 0, 1, 'Q', 'Stokes Q')
        U = self._get_parameter_from_input(U, 0, 1, 'U', 'Stokes U')
        V = self._get_parameter_from_input(V, 0, 1, 'V', 'Stokes V')

        # add the children

        self._add_child(I)
        self._add_child(Q)
        self._add_child(U)
        self._add_child(V)


#     def to_linear_polarization(self):
#         # polarization angle
# #        psi = 0.5 * np.arctan2(U_bin, Q_bin)
#
#         # polarization fraction
# #        frac = np.sqrt(Q_bin ** 2 + U_bin ** 2) / I_bin
#
#         pass
#
#         #angle = 0.5 * np.arctan2(se)
#
#



import astropy.units as astropy_units
import numpy as np
from past.utils import old_div

import astromodels.functions.numba_functions as nb_func
from astromodels.core.units import get_units
from astromodels.functions.function import (Function1D, FunctionMeta,
                                            ModelAssertionViolation)


def get_polynomial(order: int) -> Function1D:
    """
    get a polynomial function of order

    :param order: the order of the polynomical
    :type order: int
    :returns: 

    """
    return [Constant(), Line(),Quadratic(), Cubic(), Quartic()][order]
    
    
    
    


class Constant(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        Return k

    latex : $ k $

    parameters :

        k :

            desc : Constant value
            initial value : 0

    """
    def _set_units(self, x_unit, y_unit):
        self.k.unit = y_unit

    def evaluate(self, x, k):

        return k * np.ones(np.shape(x))


class Line(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A linear function

    latex : $ b * x + a $

    parameters :

        a :

            desc :  intercept
            initial value : 0

        b :

            desc : coeff
            initial value : 1

    """
    def _set_units(self, x_unit, y_unit):
        # a has units of y_unit / x_unit, so that a*x has units of y_unit
        self.a.unit = y_unit

        # b has units of y
        self.b.unit = y_unit / x_unit

    def evaluate(self, x, a, b):
        return b * x + a


class Quadratic(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A Quadratic function

    latex : $ a + b \cdot x + c \cdot x^2 $

    parameters :

        a :

            desc : coefficient
            initial value : 1

        b :

            desc : coefficient
            initial value : 1

        c :

            desc : coefficient
            initial value : 1


    """
    def _set_units(self, x_unit, y_unit):
        # a has units of y_unit / x_unit, so that a*x has units of y_unit
        self.a.unit = y_unit

        # b has units of y
        self.b.unit = y_unit / x_unit

        self.c.unit = y_unit / (x_unit) ** 2

    def evaluate(self, x, a, b, c):
        return a + b * x + c * x * x


class Cubic(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A cubic function

    latex : $ a + b \cdot x + c \cdot x^2 + d \cdot x^3$

    parameters :

        a :

            desc : coefficient
            initial value : 1

        b :

            desc : coefficient
            initial value : 1

        c :

            desc : coefficient
            initial value : 1

        d :

            desc : coefficient
            initial value : 1


    """

    def _set_units(self, x_unit, y_unit):
        # a has units of y_unit / x_unit, so that a*x has units of y_unit
        self.a.unit = y_unit

        # b has units of y
        self.b.unit = y_unit / x_unit

        self.c.unit = y_unit / (x_unit) ** 2

        self.d.unit = y_unit / (x_unit) ** 3

    def evaluate(self, x, a, b, c, d):

        x2 = x * x

        x3 = x2 * x

        return a + b * x + c * x2 + d * x3


class Quartic(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A quartic function

    latex : $ a + b \cdot x + c \cdot x^2 + d \cdot x^3 + e \cdot x^4$

    parameters :

        a :

            desc : coefficient
            initial value : 1

        b :

            desc : coefficient
            initial value : 1

        c :

            desc : coefficient
            initial value : 1

        d :

            desc : coefficient
            initial value : 1

        e :

            desc : coefficient
            initial value : 1


    """

    def _set_units(self, x_unit, y_unit):
        # a has units of y_unit / x_unit, so that a*x has units of y_unit
        self.a.unit = y_unit

        # b has units of y
        self.b.unit = y_unit / x_unit

        self.c.unit = y_unit / (x_unit) ** 2

        self.d.unit = y_unit / (x_unit) ** 3

        self.e.unit = y_unit / (x_unit) ** 4

    def evaluate(self, x, a, b, c, d, e):

        x2 = x * x

        x3 = x2 * x

        x4 = x3 * x

        return a + b * x + c * x2 + d * x3 + e * x4



from builtins import object
__author__ = 'giacomov'

import collections

import astropy.units as u

from astromodels.core.tree import Node
from astromodels.utils.pretty_list import dict_to_list

# This module keeps the configuration of the units used in astromodels

# Pre-defined values

_ENERGY = u.keV
_TIME = u.s
_ANGLE = u.deg
_AREA = u.cm**2


class UnknownUnit(Exception):
    pass


class UnitMismatch(Exception):
    pass


def _check_unit(new_unit, old_unit):
    """
    Check that the new unit is compatible with the old unit for the quantity described by variable_name

    :param new_unit: instance of astropy.units.Unit
    :param old_unit: instance of astropy.units.Unit
    :return: nothin
    """

    try:

        new_unit.physical_type

    except AttributeError:

        raise UnitMismatch("The provided unit (%s) has no physical type. Was expecting a unit for %s"
                           % (new_unit, old_unit.physical_type))

    if new_unit.physical_type != old_unit.physical_type:

        raise UnitMismatch("Physical type mismatch: you provided a unit for %s instead of a unit for %s"
                           % (new_unit.physical_type, old_unit.physical_type))


class _AstromodelsUnits(object):
    """
    Store the fundamental units of time, energy, angle and area to be used in astromodels.
    """
    def __init__(self, energy_unit=None, time_unit=None, angle_unit=None, area_unit=None):

        if energy_unit is None: energy_unit = _ENERGY
        if time_unit is None: time_unit = _TIME
        if angle_unit is None: angle_unit = _ANGLE
        if area_unit is None: area_unit = _AREA

        self._units = collections.OrderedDict()

        self._units['energy'] = energy_unit
        self._units['time'] = time_unit
        self._units['angle'] = angle_unit
        self._units['area'] = area_unit

    # This __new__ method add the properties to the class. We could have achieved the same with a metaclass,
    # but this method is more clearer, for a tiny performance penalty. Consider also that under normal circumstances
    # this class will be only created once per session

    def __new__(cls, *args, **kwargs):

        cls.energy = property(*(cls._create_property('energy')))
        cls.time = property(*(cls._create_property('time')))
        cls.angle = property(*(cls._create_property('angle')))
        cls.area = property(*(cls._create_property('area')))

        obj = super(_AstromodelsUnits, cls).__new__(cls)

        return obj

    def get(self, what):

        return self._get_unit(what)

    def _set_unit(self, what, new_unit):

        try:

            old_unit = self._units[what]

        except KeyError:

            raise UnknownUnit("You can only assign units for energy, time, angle and area. Don't know "
                              "anything about %s" % what)

        # This allows to use strings in place of Unit instances as new_unit

        new_unit = u.Unit(new_unit)

        # Check that old and new unit are for the appropriate quantity

        _check_unit(new_unit, old_unit)

        # set the new unit
        self._units[what] = new_unit

    def _get_unit(self, what):

        try:

            return self._units[what]

        except KeyError:

            raise UnknownUnit("%s is not a fundamental unit" % what)

    # This is a function which generates the elements needed to make a property.photon
    # Just a trick to avoid having to duplicate the same code for each unit.
    # It is called in __new__

    @staticmethod
    def _create_property(what):

        return (lambda self: self._get_unit(what), lambda self, new_unit: self._set_unit(what, new_unit),
                "Sets or gets the unit for %s" % what)

    # Add the == and != operators
    def __eq__(self, other):

        return other.to_dict() == self.to_dict()

    def __ne__(self, other):

        return other.to_dict() != self.to_dict()

    def _repr__base(self, rich_output):

        return dict_to_list(self._units, html=rich_output)

    def to_dict(self, minimal=False):

        return self._units


# This is a factory which will always return the same instance of the _AstromodelsUnits class

class _AstromodelsUnitsFactory(object):

        _instance = None

        def __call__(self, *args, **kwds):

            if self._instance is None:

                # Create and return a new instance

                self._instance = _AstromodelsUnits(*args, **kwds)

                return self._instance

            else:

                # Use the instance already created

                return self._instance


# Create the factory to be used in the program

get_units = _AstromodelsUnitsFactory()


from omegaconf import OmegaConf

from astromodels.utils import get_path_of_user_config

from .config_structure import Config

# Read the default Config
astromodels_config: Config = OmegaConf.structured(Config)

# now glob the config directory

for user_config_file in get_path_of_user_config().glob("*.yml"):

    _partial_conf = OmegaConf.load(user_config_file)

    astromodels_config: Config = OmegaConf.merge(astromodels_config, _partial_conf)



from __future__ import division

import astropy.units as astropy_units
import numpy as np
from past.utils import old_div

import astromodels.functions.numba_functions as nb_func
from astromodels.core.units import get_units
from astromodels.functions.function import Function1D, FunctionMeta
from astromodels.utils.configuration import astromodels_config
from astromodels.utils.logging import setup_logger

log = setup_logger(__name__)

__author__ = "giacomov"
# DMFitFunction and DMSpectra add by Andrea Albert (aalbert@slac.stanford.edu) Oct 26, 2016

erg2keV = 6.24151e8


class GSLNotAvailable(ImportWarning):
    pass


class NaimaNotAvailable(ImportWarning):
    pass



class InvalidUsageForFunction(Exception):
    pass


# Now let's try and import optional dependencies

import astropy.units as u

try:

    # Naima is for numerical computation of Synch. and Inverse compton spectra in randomly oriented
    # magnetic fields


    import naima

except ImportError:

    if astromodels_config.logging.startup_warnings:

        log.warning(
            "The naima package is not available. Models that depend on it will not be available"
        )

    has_naima = False

else:

    has_naima = True

try:

    # GSL is the GNU Scientific Library. Pygsl is the python wrapper for it. It is used by some
    # functions for faster computation

    from pygsl.testing.sf import gamma_inc

except ImportError:

    if astromodels_config.logging.startup_warnings:

        log.warning(
            "The GSL library or the pygsl wrapper cannot be loaded. Models that depend on it will not be "
            "available."
        )

    has_gsl = False

else:

    has_gsl = True



class StepFunction(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A function which is constant on the interval lower_bound - upper_bound and 0 outside the interval. The
        extremes of the interval are counted as part of the interval.

    latex : $ f(x)=\begin{cases}0 & x < \text{lower_bound} \\\text{value} & \text{lower_bound} \le x \le \text{upper_bound} \\ 0 & x > \text{upper_bound} \end{cases}$

    parameters :

        lower_bound :

            desc : Lower bound for the interval
            initial value : 0

        upper_bound :

            desc : Upper bound for the interval
            initial value : 1

        value :

            desc : Value in the interval
            initial value : 1.0

    tests :
        - { x : 0.5, function value: 1.0, tolerance: 1e-20}
        - { x : -0.5, function value: 0, tolerance: 1e-20}

    """

    def _set_units(self, x_unit, y_unit):
        # Lower and upper bound has the same unit as x
        self.lower_bound.unit = x_unit
        self.upper_bound.unit = x_unit

        # value has the same unit as y
        self.value.unit = y_unit

    def evaluate(self, x, lower_bound, upper_bound, value):
        # The value * 0 is to keep the units right

        result = np.zeros(x.shape) * value * 0

        idx = (x >= lower_bound) & (x <= upper_bound)
        result[idx] = value

        return result


class StepFunctionUpper(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A function which is constant on the interval lower_bound - upper_bound and 0 outside the interval. The
        upper interval is open.

    latex : $ f(x)=\begin{cases}0 & x < \text{lower_bound} \\\text{value} & \text{lower_bound} \le x \le \text{upper_bound} \\ 0 & x > \text{upper_bound} \end{cases}$

    parameters :

        lower_bound :

            desc : Lower bound for the interval
            initial value : 0
            fix : yes

        upper_bound :

            desc : Upper bound for the interval
            initial value : 1
            fix : yes

        value :

            desc : Value in the interval
            initial value : 1.0

    tests :
        - { x : 0.5, function value: 1.0, tolerance: 1e-20}
        - { x : -0.5, function value: 0, tolerance: 1e-20}

    """

    def _set_units(self, x_unit, y_unit):
        # Lower and upper bound has the same unit as x
        self.lower_bound.unit = x_unit
        self.upper_bound.unit = x_unit

        # value has the same unit as y
        self.value.unit = y_unit

    def evaluate(self, x, lower_bound, upper_bound, value):
        # The value * 0 is to keep the units right

        result = np.zeros(x.shape) * value * 0

        idx = (x >= lower_bound) & (x < upper_bound)
        result[idx] = value

        return result


# noinspection PyPep8Naming


class Blackbody(Function1D, metaclass=FunctionMeta):
    r"""

    description :
        A blackbody function

    latex : $f(x) = K \frac{x^2}{\exp(\frac{x}{kT}) -1}  $

    parameters :
        K :
            desc :
            initial value : 1e-4
            min : 0.
            is_normalization : True

        kT :
            desc : temperature of the blackbody
            initial value : 30.0
            min: 0.

    """

    def _set_units(self, x_unit, y_unit):
        # The normalization has the same units as y
        self.K.unit = old_div(y_unit, (x_unit ** 2))

        # The break point has always the same dimension as the x variable
        self.kT.unit = x_unit

    def evaluate(self, x, K, kT):

        if isinstance(x, astropy_units.Quantity):

            K_ = K.value
            kT_ = kT.value

            x_ = x.value

            unit_ = self.y_unit

        else:
            unit_ = 1.0
            K_, kT_, x_, = (
                K,
                kT,
                x,
            )

        result = nb_func.bb_eval(x_, K_, kT_)

        return result * unit_


# noinspection PyPep8Naming


class Sin(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A sinusodial function

    latex : $ K~\sin{(2\pi f x + \phi)} $

    parameters :

        K :

            desc : Normalization
            initial value : 1
            is_normalization : True

        f :

            desc : frequency
            initial value : 1.0 / (2 * np.pi)
            min : 0

        phi :

            desc : phase
            initial value : 0
            min : -np.pi
            max : +np.pi
            unit: rad

    tests :
        - { x : 0.0, function value: 0.0, tolerance: 1e-10}
        - { x : 1.5707963267948966, function value: 1.0, tolerance: 1e-10}

    """

    def _set_units(self, x_unit, y_unit):
        # The normalization has the same unit of y
        self.K.unit = y_unit

        # The unit of f is 1 / [x] because fx must be a pure number. However,
        # np.pi of course doesn't have units, so we add a rad
        self.f.unit = x_unit ** (-1) * astropy_units.rad

        # The unit of phi is always the same (radians)

        self.phi.unit = astropy_units.rad

    # noinspection PyPep8Naming
    def evaluate(self, x, K, f, phi):
        return K * np.sin(2 * np.pi * f * x + phi)




class DiracDelta(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        return  at zero_point

    latex : $ value $

    parameters :

        value :

            desc : Constant value
            initial value : 0

        zero_point:

             desc: value at which function is non-zero
             initial value : 0
             fix : yes


    """

    def _set_units(self, x_unit, y_unit):

        self.value.unit = y_unit
        self.zero_point.unit = x_unit

    def evaluate(self, x, value, zero_point):

        out = np.zeros(x.shape) * value * 0

        out[x == zero_point] = value

        return out


if has_naima:

    class Synchrotron(Function1D, metaclass=FunctionMeta):
        r"""
        description :
            Synchrotron spectrum from an input particle distribution, using Naima (naima.readthedocs.org)
        latex: not available
        parameters :
            B :
                desc : magnetic field
                initial value : 3.24e-6
                unit: Gauss
            distance :
                desc : distance of the source
                initial value : 1.0
                unit : kpc
            emin :
                desc : minimum energy for the particle distribution
                initial value : 1
                fix : yes
                unit: GeV
            emax :
                desc : maximum energy for the particle distribution
                initial value : 510e3
                fix : yes
                unit: GeV
            need:
                desc: number of points per decade in which to evaluate the function
                initial value : 10
                min : 2
                max : 100
                fix : yes
        """

        def _set_units(self, x_unit, y_unit):

            # This function can only be used as a spectrum,
            # so let's check that x_unit is a energy and y_unit is
            # differential flux

            if hasattr(x_unit, "physical_type") and x_unit.physical_type == "energy":

                # Now check that y is a differential flux
                current_units = get_units()
                should_be_unitless = y_unit * (
                    current_units.energy * current_units.time * current_units.area
                )

                if (
                    not hasattr(should_be_unitless, "physical_type")
                    or should_be_unitless.decompose().physical_type != "dimensionless"
                ):
                    # y is not a differential flux
                    raise InvalidUsageForFunction(
                        "Unit for y is not differential flux. The function synchrotron "
                        "can only be used as a spectrum."
                    )
            else:

                raise InvalidUsageForFunction(
                    "Unit for x is not an energy. The function synchrotron can only be used "
                    "as a spectrum"
                )

                # we actually don't need to do anything as the units are already set up

        def set_particle_distribution(self, function):

            self._particle_distribution = function

            # Now set the units for the function

            current_units = get_units()

            self._particle_distribution.set_units(
                current_units.energy, current_units.energy ** (-1)
            )

            # Naima wants a function which accepts a quantity as x (in units of eV) and returns an astropy quantity,
            # so we need to create a wrapper which will remove the unit from x and add the unit to the return
            # value

            self._particle_distribution_wrapper = lambda x: old_div(
                function(x.value), current_units.energy
            )

        def get_particle_distribution(self):

            return self._particle_distribution

        particle_distribution = property(
            get_particle_distribution,
            set_particle_distribution,
            doc="""Get/set particle distribution for electrons""",
        )

        def fix_units(self, x, B, distance, emin, emax):

            if isinstance(x, u.Quantity):
                return (
                    True,
                    x.to(get_units().energy),
                    B.to(u.Gauss),
                    distance.to(u.kpc),
                    emin.to(u.GeV),
                    emax.to(u.GeV),
                )
            else:
                return (
                    False,
                    x * (get_units().energy),
                    B * (u.Gauss),
                    distance * (u.kpc),
                    emin * (u.GeV),
                    emax * (u.GeV),
                )

        # noinspection PyPep8Naming
        def evaluate(self, x, B, distance, emin, emax, need):

            has_units, x, B, distance, emin, emax = self.fix_units(
                x, B, distance, emin, emax
            )

            _synch = naima.models.Synchrotron(
                self._particle_distribution_wrapper,
                B,
                Eemin=emin,
                Eemax=emax,
                nEed=need,
            )

            if has_units:
                return _synch.flux(x, distance=distance)
            else:
                return _synch.flux(x, distance=distance).value

        def to_dict(self, minimal=False):

            data = super(Function1D, self).to_dict(minimal)

            if not minimal:
                data["extra_setup"] = {
                    "particle_distribution": self.particle_distribution.path
                }

            return data


class _ComplexTestFunction(Function1D, metaclass=FunctionMeta):
    r"""
    description :
        A useless function to be used during automatic tests

    latex: not available

    parameters :
        A :
            desc : none
            initial value : 3.24e-6
            min : 1e-6
            max : 1e-5

        B :
            desc : none
            initial value : -10
            min : -100
            max : 100
            delta : 0.1
    """

    def _set_units(self, x_unit, y_unit):

        self.A.unit = y_unit
        self.B.unit = old_div(y_unit, x_unit)

    def set_particle_distribution(self, function):

        self._particle_distribution = function

    def get_particle_distribution(self):

        return self._particle_distribution

    particle_distribution = property(
        get_particle_distribution,
        set_particle_distribution,
        doc="""Get/set particle distribution for electrons""",
    )

    # noinspection PyPep8Naming
    def evaluate(self, x, A, B):

        return A + B * x

    def to_dict(self, minimal=False):

        data = super(Function1D, self).to_dict(minimal)

        if not minimal:

            data["extra_setup"] = {
                "particle_distribution": self.particle_distribution.path
            }

        return data


class Log_parabola(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A log-parabolic function. NOTE that we use the high-energy convention of using the natural log in place of the
        base-10 logarithm. This means that beta is a factor 1 / log10(e) larger than what returned by those software
        using the other convention.

    latex : $ K \left( \frac{x}{piv} \right)^{\alpha -\beta \log{\left( \frac{x}{piv} \right)}} $

    parameters :

        K :

            desc : Normalization
            initial value : 1.0
            is_normalization : True
            transformation : log10
            min : 1e-30
            max : 1e5

        piv :
            desc : Pivot (keep this fixed)
            initial value : 1
            fix : yes

        alpha :

            desc : index
            initial value : -2.0

        beta :

            desc : curvature (positive is concave, negative is convex)
            initial value : 1.0

    """

    def _set_units(self, x_unit, y_unit):

        # K has units of y

        self.K.unit = y_unit

        # piv has the same dimension as x
        self.piv.unit = x_unit

        # alpha and beta are dimensionless
        self.alpha.unit = astropy_units.dimensionless_unscaled
        self.beta.unit = astropy_units.dimensionless_unscaled

    def evaluate(self, x, K, piv, alpha, beta):

        # print("Receiving %s" % ([K, piv, alpha, beta]))

        xx = np.divide(x, piv)

        try:

            return K * xx ** (alpha - beta * np.log(xx))

        except ValueError:

            # The current version of astropy (1.1.x) has a bug for which quantities that have become
            # dimensionless because of a division (like xx here) are not recognized as such by the power
            # operator, which throws an exception: ValueError: Quantities and Units may only be raised to a scalar power
            # This is a quick fix, waiting for astropy 1.2 which will fix this

            xx = xx.to("")

            return K * xx ** (alpha - beta * np.log(xx))

    @property
    def peak_energy(self):
        """
        Returns the peak energy in the nuFnu spectrum

        :return: peak energy in keV
        """

        # Eq. 6 in Massaro et al. 2004
        # (http://adsabs.harvard.edu/abs/2004A%26A...413..489M)

        return self.piv.value * pow(
            10, old_div(((2 + self.alpha.value) * np.log(10)),
                        (2 * self.beta.value))
        )


if has_gsl:

    class Cutoff_powerlaw_flux(Function1D, metaclass=FunctionMeta):
        r"""
        description :

            A cutoff power law having the flux as normalization, which should reduce the correlation among
            parameters.

        latex : $ \frac{F}{T(b)-T(a)} ~x^{index}~\exp{(-x/x_{c})}~\text{with}~T(x)=-x_{c}^{index+1} \Gamma(index+1, x/C)~\text{(}\Gamma\text{ is the incomplete gamma function)} $

        parameters :

            F :

                desc : Integral between a and b
                initial value : 1e-5
                is_normalization : True

            index :

                desc : photon index
                initial value : -2.0

            xc :

                desc : cutoff position
                initial value : 50.0

            a :

                desc : lower bound for the band in which computing the integral F
                initial value : 1.0
                fix : yes

            b :

                desc : upper bound for the band in which computing the integral F
                initial value : 100.0
                fix : yes
        """

        def _set_units(self, x_unit, y_unit):
            # K has units of y * x
            self.F.unit = y_unit * x_unit

            # alpha is dimensionless
            self.index.unit = astropy_units.dimensionless_unscaled

            # xc, a and b have the same dimension as x
            self.xc.unit = x_unit
            self.a.unit = x_unit
            self.b.unit = x_unit

        @staticmethod
        def _integral(a, b, index, ec):
            ap1 = index + 1

            def integrand(x):
                return -pow(ec, ap1) * gamma_inc(ap1, old_div(x, ec))

            return integrand(b) - integrand(a)

        def evaluate(self, x, F, index, xc, a, b):
            this_integral = self._integral(a, b, index, xc)

            return (
                F / this_integral *
                np.power(x, index) * np.exp(-1 * np.divide(x, xc))
            )


class Exponential_cutoff(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        An exponential cutoff

    latex : $ K \exp{(-x/xc)} $

    parameters :

        K :

            desc : Normalization
            initial value : 1.0
            fix : no
            is_normalization : True

        xc :
            desc : cutoff
            initial value : 100
            min : 1
    """

    def _set_units(self, x_unit, y_unit):
        # K has units of y

        self.K.unit = y_unit

        # piv has the same dimension as x
        self.xc.unit = x_unit

    def evaluate(self, x, K, xc):
        return K * np.exp(np.divide(x, -xc))



#!/usr/bin/env python

from __future__ import print_function
import ctypes.util
import glob
import sys

import os
import re
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext

import versioneer

# This is needed to use numpy in this module, and should work whether or not numpy is
# already installed. If it's not, it will trigger an installation

class My_build_ext(_build_ext):

    def finalize_options(self):

        _build_ext.finalize_options(self)

        # Prevent numpy from thinking it is still in its setup process:

        __builtins__.__NUMPY_SETUP__ = False

        import numpy

        self.include_dirs.append(numpy.get_include())
        self.include_dirs.append('astromodels/xspec/include')


def sanitize_lib_name(library_path):
    """
    Get a fully-qualified library name, like /usr/lib/libgfortran.so.3.0, and returns the lib name needed to be
    passed to the linker in the -l option (for example gfortran)

    :param library_path:
    :return:
    """

    lib_name = os.path.basename(library_path)

    # Some regexp magic needed to extract in a system-independent (mac/linux) way the library name

    tokens = re.findall("lib(.+)(\.so|\.dylib|\.a)(.+)?", lib_name)

    if not tokens:
        raise RuntimeError('Attempting to find %s in directory %s but there are no libraries in this directory'%(lib_name,library_path))


    return tokens[0][0]


def find_library(library_root, additional_places=None):
    """
    Returns the name of the library without extension

    :param library_root: root of the library to search, for example "cfitsio_" will match libcfitsio_1.2.3.4.so
    :return: the name of the library found (NOTE: this is *not* the path), and a directory path if the library is not
    in the system paths (and None otherwise). The name of libcfitsio_1.2.3.4.so will be cfitsio_1.2.3.4, in other words,
    it will be what is needed to be passed to the linker during a c/c++ compilation, in the -l option
    """

    # find_library searches for all system paths in a system independent way (but NOT those defined in
    # LD_LIBRARY_PATH or DYLD_LIBRARY_PATH)

    first_guess = ctypes.util.find_library(library_root)

    if first_guess is not None:

        # Found in one of the system paths

        if sys.platform.lower().find("linux") >= 0:

            # On linux the linker already knows about these paths, so we
            # can return None as path

            return sanitize_lib_name(first_guess), None

        elif sys.platform.lower().find("darwin") >= 0:

            # On Mac we still need to return the path, because the linker sometimes
            # does not look into it

            return sanitize_lib_name(first_guess), os.path.dirname(first_guess)

        else:

            # Windows is not supported

            raise NotImplementedError("Platform %s is not supported" % sys.platform)

    else:

        # could not find it. Let's examine LD_LIBRARY_PATH or DYLD_LIBRARY_PATH
        # (if they sanitize_lib_name(first_guess), are not defined, possible_locations will become [""] which will
        # be handled by the next loop)

        if sys.platform.lower().find("linux") >= 0:

            # Unix / linux

            possible_locations = os.environ.get("LD_LIBRARY_PATH", "").split(":")

        elif sys.platform.lower().find("darwin") >= 0:

            # Mac

            possible_locations = os.environ.get("DYLD_LIBRARY_PATH", "").split(":")

        else:

            raise NotImplementedError("Platform %s is not supported" % sys.platform)

        if additional_places is not None:

            possible_locations.extend(additional_places)

        # Now look into the search paths

        library_name = None
        library_dir = None

        for search_path in possible_locations:

            if search_path == "":
                # This can happen if there are more than one :, or if nor LD_LIBRARY_PATH
                # nor DYLD_LIBRARY_PATH are defined (because of the default use above for os.environ.get)

                continue

            results = glob.glob(os.path.join(search_path, "lib%s*" % library_root))

            if len(results) >= 1:

                # Results contain things like libXS.so, libXSPlot.so, libXSpippo.so
                # If we are looking for libXS.so, we need to make sure that we get the right one!

                for result in results:

                    if re.match("lib%s[\-_\.]" % library_root, os.path.basename(result)) is None:

                        continue

                    else:

                        # FOUND IT

                        # This is the full path of the library, like /usr/lib/libcfitsio_1.2.3.4

                        library_name = result
                        library_dir = search_path

                        break

            else:

                continue

            if library_name is not None:
                break

        if library_name is None:

            return None, None

        else:

            # Sanitize the library name to get from the fully-qualified path to just the library name
            # (/usr/lib/libgfortran.so.3.0 becomes gfortran)

            return sanitize_lib_name(library_name), library_dir




def setup_xspec():

    headas_root = os.environ.get("HEADAS")
    conda_prefix = os.environ.get("CONDA_PREFIX")

    if headas_root is None:

        # See, maybe we are running in Conda
        
        if conda_prefix is None:
            
            # Maybe this is a Conda build
            
            conda_prefix = os.environ.get("PREFIX")

        if conda_prefix is not None:

            # Yes, this is Conda
            # Let's see if the package xspec-modelsonly has been installed by checking whether one of the Xspec
            # libraries exists within conda
            conda_lib_path = os.path.join(conda_prefix, 'lib')
            this_lib, this_lib_path = find_library('XSFunctions', additional_places=[conda_lib_path])

            if this_lib is None:

                # No, there is no library in Conda
                print("No xspec-modelsonly package has been installed in Conda. Xspec support will not be installed")

                print("Was looking into %s" % conda_lib_path)

                return None

            else:

                print("The xspec-modelsonly package has been installed in Conda. Xspec support will be installed")

                # Set up the HEADAS variable so that the following will find the libraries
                headas_root = conda_prefix

        else:

            print("No HEADAS env. variable set. Xspec support will not be installed ")

            return None

    else:

        print("\n Xspec is detected. Will compile the Xspec extension.\n")
        print("\n NOTICE!!!!!\n")
        print("If you have issues, manually set the ENV variable XSPEC_INC_PATH")
        print("To the location of the XSPEC headers\n\n")
        print("If you are still having issues, unset HEADAS before installing and contact the support team")
        


    # Make sure these libraries exist and are linkable right now
    # (they need to be in LD_LIBRARY_PATH or DYLD_LIBRARY_PATH or in one of the system paths)
    
    libraries_root = ['XSFunctions', 'XSModel', 'XSUtil', 'XS', 'cfitsio', 'CCfits', 'wcs', 'gfortran']
            
    libraries = []
    library_dirs = []

    for lib_root in libraries_root:

        this_library, this_library_path = find_library(lib_root, additional_places=[os.path.join(headas_root, 'lib')])

        if this_library is None:

            raise IOError("Could not find library %s. Impossible to compile Xspec" % lib_root)

        else:

            print("Found library %s in %s" % (this_library, this_library_path))

            libraries.append(this_library)

            if this_library_path is not None:
                # This library is not in one of the system path library, we need to add
                # it to the -L flag during linking. Let's put it in the library_dirs list
                # which will be used in the Extension class

                library_dirs.append(this_library_path)


    # try to manually add on the include directory

    header_paths = []
    
    if library_dirs:

        # grab it from the lib assuming that it is one up
        xspec_path, _ = os.path.split(library_dirs[0])
        include_path = os.path.join(xspec_path, "include")

        header_paths.append(include_path)

    # let's be sure to add the conda include directory
       
    if conda_prefix is not None:
        
        conda_include_path = os.path.join(conda_prefix, 'include')
        header_paths.append(conda_include_path)

    # check if there are user set the location of the xspec headers:
   
    xspec_headers_path = os.environ.get("XSPEC_INC_PATH")
    
    if xspec_headers_path is not None:

        print("You have set XSPEC_INC_PATH=%s" % xspec_headers_path)

        header_paths.append(xspec_headers_path)
    
    # Remove duplicates from library_dirs and header_paths

    library_dirs = list(set(library_dirs))  
    header_paths = list(set(header_paths))

    # Configure the variables to build the external module with the C/C++ wrapper


    ext_modules_configuration = [

        Extension("astromodels.xspec._xspec",

                  ["astromodels/xspec/src/_xspec.cc", ],
                  include_dirs=header_paths,
                  libraries=libraries,
                  library_dirs=library_dirs,
                  runtime_library_dirs=library_dirs,
                  extra_compile_args=[])]

    return ext_modules_configuration


# Normal packages

packages = ['astromodels',
            'astromodels/core',
            'astromodels/functions',
            'astromodels/functions/functions_1D',
            'astromodels/functions/dark_matter',
            'astromodels/sources',
            'astromodels/utils',
            'astromodels/xspec',
            'astromodels/tests'
            ]

# Check whether we can compile Xspec support
ext_modules_configuration = setup_xspec()

# Add the node_ctype module

# This defines the external module
node_ctype_ext = Extension('astromodels.core.node_ctype',
                           sources = ['astromodels/core/node_ctype/node_ctype.cxx'],
                           extra_compile_args=[]) # '-UNDEBUG' for debugging


if ext_modules_configuration is None:

    # No Xspec
    ext_modules_configuration = [node_ctype_ext]

else:

    ext_modules_configuration.append(node_ctype_ext)


setup(

    setup_requires=['numpy'],

    #cmdclass={'build_ext': My_build_ext},
    cmdclass=versioneer.get_cmdclass({'build_ext': My_build_ext}),
    
    packages=packages,

    data_files=[('astromodels/data/functions', glob.glob('astromodels/data/functions/*.yaml')),
                ('astromodels/data/tests',  glob.glob('astromodels/data/tests/*.fits'))

    ],

    # The __version__ comes from the exec at the top

    version=versioneer.get_version(),


    download_url='https://github.com/threeml/astromodels/archive/v0.1',

    keywords=['Likelihood', 'Models', 'fit'],
    

    ext_modules=ext_modules_configuration,

    package_data={
        'astromodels': ['data/dark_matter/*', 'data/xsect/*', 'data/past_1D_values.h5'],
    },


)
